#include "ba0.h"

static double
invert_double (
    double z,
    char **mesg)
{
  struct ba0_mark M;

  ba0_record (&M);
// simulate some temporary memory allocation
  ba0_alloc (10);
  if (z == 0.)
    {
// the string is allocated in the main stack
      *mesg = ba0_strdup ("zero is not invertible!");
      BA0_RAISE_EXCEPTION2 (BA0_ERRIVZ, "%s", mesg);
    }
  else if (ba0_isnan (z))
    {
      ba0_push_another_stack ();
// the second stack becomes the current stack
// the string is allocated in the second stack
      *mesg = ba0_strdup ("nan is not invertible!");
      BA0_RAISE_EXCEPTION2 (BA0_ERRIVZ, "%s", mesg);
    }
// the temporary area is freed only if no exception was raised 
  ba0_restore (&M);
  return 1. / z;
}

int
main (
    void)
{
  struct ba0_arrayof_double T;
  struct ba0_stack *H;
  struct ba0_mark M;
  bool b = true;

  ba0_restart (0, 0);
  ba0_init_array ((struct ba0_array *) &T);
  ba0_sscanf2 ("[3., 0., nan, -5]", "%a[%le]", &T);
  H = ba0_current_stack ();
  ba0_record (&M);
  for (int i = 0; i < T.size; i++)
    {
      char *mesg;
      BA0_TRY
      {
        double d;
        d = invert_double (T.tab[i], &mesg);
        ba0_printf ("the inverse of %le is: %le\n", &T.tab[i], &d);
// Check that the call is successful for i = 0 or 3
        b = b && (i == 0 || i == 3);
      }
      BA0_CATCH
      {
        if (ba0_global.exception.raised != BA0_ERRIVZ)
          BA0_RE_RAISE_EXCEPTION;
        ba0_printf ("%s\n", mesg);
// Check that exception BA0_ERRIVZ was raised for i == 1 or 2
//   and that mesg is in the main stack
        b = b && (i == 1 || i == 2) && ba0_in_stack (mesg, H);
      }
      BA0_ENDTRY;
    }
// Check that the current stack is the main one
  b = b && H == ba0_current_stack ();
  ba0_terminate (ba0_init_level);
  return b ? 0 : 1;
}
