#include "bap_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_itercoeff_COEFF.h"
#include "bap__check_COEFF.h"
#include "bap_add_polynom_COEFF.h"

#define BAD_FLAG_COEFF

/****************************************************************************
 ADDITION
 ****************************************************************************/

/*
 * texinfo: bap_add_polynom_COEFF
 * Assign @math{A + B} to @var{R}.
 */

BAP_DLL void
bap_add_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bap_polynom_COEFF *B)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t bunk;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_set_term (&TB, &B->total_rank);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    bap_term_itermon_COEFF (&TA, &iterA);
  if (!outB)
    bap_term_itermon_COEFF (&TB, &iterB);
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
          bap_write_creator_COEFF (&crea, &TB,
              *bap_coeff_itermon_COEFF (&iterB));
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          break;
        case ba0_gt:
          bap_write_creator_COEFF (&crea, &TA,
              *bap_coeff_itermon_COEFF (&iterA));
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
          break;
        default:
          COEFF_add (bunk, *bap_coeff_itermon_COEFF (&iterA),
              *bap_coeff_itermon_COEFF (&iterB));
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
        }
    }
  while (!outA)
    {
      bap_write_creator_COEFF (&crea, &TA, *bap_coeff_itermon_COEFF (&iterA));
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        bap_term_itermon_COEFF (&TA, &iterA);
    }
  while (!outB)
    {
      bap_write_creator_COEFF (&crea, &TB, *bap_coeff_itermon_COEFF (&iterB));
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        bap_term_itermon_COEFF (&TB, &iterB);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * texinfo: bap_add_polynom_numeric_COEFF
 * Assign @math{A + c} to @var{R}.
 */

BAP_DLL void
bap_add_polynom_numeric_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    COEFF_t c)
{
  struct ba0_mark M;
  struct bav_rank rg;

  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (bap_is_zero_polynom_COEFF (A))
    {
      rg = bav_constant_rank ();
      bap_set_polynom_crk_COEFF (R, c, &rg);
    }
  else if (R == A)
    {
      if (!COEFF_is_zero (c))
        {
          struct bav_term T;
          struct bap_itermon_COEFF iter;
          COEFF_t *lc;

          ba0_push_another_stack ();
          ba0_record (&M);
          bav_init_term (&T);
          ba0_pull_stack ();
          bap_end_itermon_COEFF (&iter, R);
          ba0_push_another_stack ();
          bap_term_itermon_COEFF (&T, &iter);
          ba0_pull_stack ();
          if (bav_is_one_term (&T))
            {
              lc = bap_coeff_itermon_COEFF (&iter);
              COEFF_add (*lc, *lc, c);
              if (COEFF_is_zero (*lc))
                {
                  if (R->access == bap_sequential_monom_access)
                    R->seq.after--;
                  else
                    R->ind.size--;
                }
            }
          else
            {
              struct bap_creator_COEFF crea;

              bap_append_creator_COEFF (&crea, R, 1);
              bav_set_term_one (&T);
              bap_write_creator_COEFF (&crea, &T, c);
              bap_close_creator_COEFF (&crea);
            }
          ba0_restore (&M);
        }
    }
  else if (!COEFF_is_zero (c))
    {
      struct bap_polynom_COEFF *P;
      struct bav_rank rg;

      ba0_push_another_stack ();
      ba0_record (&M);
      rg = bav_constant_rank ();
      P = bap_new_polynom_crk_COEFF (c, &rg);
      ba0_pull_stack ();
      bap_add_polynom_COEFF (R, A, P);
      ba0_restore (&M);
    }
  else
    bap_set_polynom_COEFF (R, A);
}

/*
 * texinfo: bap_sub_polynom_COEFF
 * Assign @math{A - B} to @var{R}.
 */

BAP_DLL void
bap_sub_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bap_polynom_COEFF *B)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t bunk;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_set_term (&TB, &B->total_rank);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    bap_term_itermon_COEFF (&TA, &iterA);
  if (!outB)
    bap_term_itermon_COEFF (&TB, &iterB);
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
          bap_write_neg_creator_COEFF (&crea, &TB,
              *bap_coeff_itermon_COEFF (&iterB));
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          break;
        case ba0_gt:
          bap_write_creator_COEFF (&crea, &TA,
              *bap_coeff_itermon_COEFF (&iterA));
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
          break;
        default:
          COEFF_sub (bunk, *bap_coeff_itermon_COEFF (&iterA),
              *bap_coeff_itermon_COEFF (&iterB));
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
        }
    }
  while (!outA)
    {
      bap_write_creator_COEFF (&crea, &TA, *bap_coeff_itermon_COEFF (&iterA));
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        bap_term_itermon_COEFF (&TA, &iterA);
    }
  while (!outB)
    {
      bap_write_neg_creator_COEFF (&crea, &TB,
          *bap_coeff_itermon_COEFF (&iterB));
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        bap_term_itermon_COEFF (&TB, &iterB);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * R = A - q * TQ * B
 */

/*
 * texinfo: bap_submulmon_polynom_COEFF
 * Assign @math{A - q\,@emph{TQ}\,B} to @var{R}.
 */

BAP_DLL void
bap_submulmon_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bap_polynom_COEFF *B,
    struct bav_term *TQ,
    COEFF_t q)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t cB, bunk;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);
  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_set_term (&TB, &B->total_rank);
  bav_mul_term (&TB, &TB, TQ);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  COEFF_init (cB);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    bap_term_itermon_COEFF (&TA, &iterA);
  if (!outB)
    {
      bap_term_itermon_COEFF (&TB, &iterB);
      bav_mul_term (&TB, &TB, TQ);
      COEFF_mul (cB, *bap_coeff_itermon_COEFF (&iterB), q);
    }
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
          bap_write_neg_creator_COEFF (&crea, &TB, cB);
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            {
              bap_term_itermon_COEFF (&TB, &iterB);
              bav_mul_term (&TB, &TB, TQ);
              COEFF_mul (cB, *bap_coeff_itermon_COEFF (&iterB), q);
            }
          break;
        case ba0_gt:
          bap_write_creator_COEFF (&crea, &TA,
              *bap_coeff_itermon_COEFF (&iterA));
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
          break;
        default:
          COEFF_sub (bunk, *bap_coeff_itermon_COEFF (&iterA), cB);
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            {
              bap_term_itermon_COEFF (&TB, &iterB);
              bav_mul_term (&TB, &TB, TQ);
              COEFF_mul (cB, *bap_coeff_itermon_COEFF (&iterB), q);
            }
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
        }
    }
  while (!outA)
    {
      bap_write_creator_COEFF (&crea, &TA, *bap_coeff_itermon_COEFF (&iterA));
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        bap_term_itermon_COEFF (&TA, &iterA);
    }
  while (!outB)
    {
      bap_write_neg_creator_COEFF (&crea, &TB, cB);
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        {
          bap_term_itermon_COEFF (&TB, &iterB);
          bav_mul_term (&TB, &TB, TQ);
          COEFF_mul (cB, *bap_coeff_itermon_COEFF (&iterB), q);
        }
    }

  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * R = cA * A + cB * B
 */

/*
 * texinfo: bap_comblin_polynom_COEFF
 * Assign the linear combination @math{@emph{cA}\,A + @emph{cB}\,B}
 * to @var{R}.
 */

BAP_DLL void
bap_comblin_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    ba0_int_p cA,
    struct bap_polynom_COEFF *B,
    ba0_int_p cB)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t bunk, bink;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (cA == 0 || cB == 0)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_set_term (&TB, &B->total_rank);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  COEFF_init (bink);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    bap_term_itermon_COEFF (&TA, &iterA);
  if (!outB)
    bap_term_itermon_COEFF (&TB, &iterB);
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
/*
 * Specific cast.
 * GMP function expect long which is 32 bits instead of 64 on Windows 64.
 */
          COEFF_mul_si (bunk, *bap_coeff_itermon_COEFF (&iterB), (long) cB);
          bap_write_creator_COEFF (&crea, &TB, bunk);
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          break;
        case ba0_gt:
/*
 * Specific cast.
 * GMP function expect long which is 32 bits instead of 64 on Windows 64.
 */
          COEFF_mul_si (bunk, *bap_coeff_itermon_COEFF (&iterA), (long) cA);
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
          break;
        default:
/*
 * Specific cast.
 * GMP function expect long which is 32 bits instead of 64 on Windows 64.
 */
          COEFF_mul_si (bunk, *bap_coeff_itermon_COEFF (&iterA), (long) cA);
          COEFF_mul_si (bink, *bap_coeff_itermon_COEFF (&iterB), (long) cB);
          COEFF_add (bunk, bunk, bink);
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          if (!outA)
            bap_term_itermon_COEFF (&TA, &iterA);
        }
    }
  while (!outA)
/*
 * Specific cast.
 * GMP function expect long which is 32 bits instead of 64 on Windows 64.
 */
    {
      COEFF_mul_si (bunk, *bap_coeff_itermon_COEFF (&iterA), (long) cA);
      bap_write_creator_COEFF (&crea, &TA, bunk);
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        bap_term_itermon_COEFF (&TA, &iterA);
    }
  while (!outB)
/*
 * Specific cast.
 * GMP function expect long which is 32 bits instead of 64 on Windows 64.
 */
    {
      COEFF_mul_si (bunk, *bap_coeff_itermon_COEFF (&iterB), (long) cB);
      bap_write_creator_COEFF (&crea, &TB, bunk);
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        bap_term_itermon_COEFF (&TB, &iterB);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * R = A * rg + B
 */

/*
 * texinfo: bap_addmulrk_polynom_COEFF
 * Assign @math{A\,@emph{rg} + B} to @var{R}.
 */

BAP_DLL void
bap_addmulrk_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bav_rank *rg,
    struct bap_polynom_COEFF *B)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t bunk;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_mul_term_rank (&TA, &TA, rg);
  bav_set_term (&TB, &B->total_rank);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    {
      bap_term_itermon_COEFF (&TA, &iterA);
      bav_mul_term_rank (&TA, &TA, rg);
    }
  if (!outB)
    bap_term_itermon_COEFF (&TB, &iterB);
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
          bap_write_creator_COEFF (&crea, &TB,
              *bap_coeff_itermon_COEFF (&iterB));
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          break;
        case ba0_gt:
          bap_write_creator_COEFF (&crea, &TA,
              *bap_coeff_itermon_COEFF (&iterA));
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            {
              bap_term_itermon_COEFF (&TA, &iterA);
              bav_mul_term_rank (&TA, &TA, rg);
            }
          break;
        default:
          COEFF_add (bunk, *bap_coeff_itermon_COEFF (&iterA),
              *bap_coeff_itermon_COEFF (&iterB));
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          if (!outA)
            {
              bap_term_itermon_COEFF (&TA, &iterA);
              bav_mul_term_rank (&TA, &TA, rg);
            }
        }
    }
  while (!outA)
    {
      bap_write_creator_COEFF (&crea, &TA, *bap_coeff_itermon_COEFF (&iterA));
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        {
          bap_term_itermon_COEFF (&TA, &iterA);
          bav_mul_term_rank (&TA, &TA, rg);
        }
    }
  while (!outB)
    {
      bap_write_creator_COEFF (&crea, &TB, *bap_coeff_itermon_COEFF (&iterB));
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        bap_term_itermon_COEFF (&TB, &iterB);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * R = A * rg - B
 */

/*
 * texinfo: bap_submulrk_polynom_COEFF
 * Assign @math{A\,@emph{rg} - B} to @var{R}.
 */

BAP_DLL void
bap_submulrk_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bav_rank *rg,
    struct bap_polynom_COEFF *B)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF *P;
  enum ba0_compare_code code;
  struct bav_term TA, TB;
  COEFF_t bunk;
  bool outA, outB;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_set_term (&TA, &A->total_rank);
  bav_mul_term_rank (&TA, &TA, rg);
  bav_set_term (&TB, &B->total_rank);
  bav_lcm_term (&TA, &TA, &TB);

  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &TA, bap_approx_total_rank,
      BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B)));

  bap_begin_itermon_COEFF (&iterA, A);
  bap_begin_itermon_COEFF (&iterB, B);
  COEFF_init (bunk);
  outA = bap_outof_itermon_COEFF (&iterA);
  outB = bap_outof_itermon_COEFF (&iterB);
  if (!outA)
    {
      bap_term_itermon_COEFF (&TA, &iterA);
      bav_mul_term_rank (&TA, &TA, rg);
    }
  if (!outB)
    bap_term_itermon_COEFF (&TB, &iterB);
  while (!outA && !outB)
    {
      code = bav_compare_term (&TA, &TB);
      switch (code)
        {
        case ba0_lt:
          bap_write_neg_creator_COEFF (&crea, &TB,
              *bap_coeff_itermon_COEFF (&iterB));
          bap_next_itermon_COEFF (&iterB);
          outB = bap_outof_itermon_COEFF (&iterB);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          break;
        case ba0_gt:
          bap_write_creator_COEFF (&crea, &TA,
              *bap_coeff_itermon_COEFF (&iterA));
          bap_next_itermon_COEFF (&iterA);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outA)
            {
              bap_term_itermon_COEFF (&TA, &iterA);
              bav_mul_term_rank (&TA, &TA, rg);
            }
          break;
        default:
          COEFF_sub (bunk, *bap_coeff_itermon_COEFF (&iterA),
              *bap_coeff_itermon_COEFF (&iterB));
          bap_write_creator_COEFF (&crea, &TA, bunk);
          bap_next_itermon_COEFF (&iterB);
          bap_next_itermon_COEFF (&iterA);
          outB = bap_outof_itermon_COEFF (&iterB);
          outA = bap_outof_itermon_COEFF (&iterA);
          if (!outB)
            bap_term_itermon_COEFF (&TB, &iterB);
          if (!outA)
            {
              bap_term_itermon_COEFF (&TA, &iterA);
              bav_mul_term_rank (&TA, &TA, rg);
            }
        }
    }
  while (!outA)
    {
      bap_write_creator_COEFF (&crea, &TA, *bap_coeff_itermon_COEFF (&iterA));
      bap_next_itermon_COEFF (&iterA);
      outA = bap_outof_itermon_COEFF (&iterA);
      if (!outA)
        {
          bap_term_itermon_COEFF (&TA, &iterA);
          bav_mul_term_rank (&TA, &TA, rg);
        }
    }
  while (!outB)
    {
      bap_write_neg_creator_COEFF (&crea, &TB,
          *bap_coeff_itermon_COEFF (&iterB));
      bap_next_itermon_COEFF (&iterB);
      outB = bap_outof_itermon_COEFF (&iterB);
      if (!outB)
        bap_term_itermon_COEFF (&TB, &iterB);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}


#undef BAD_FLAG_COEFF
