#include "bap_polynom_COEFF.h"
#include "bap_creator_COEFF.h"
#include "bap_itermon_COEFF.h"
#include "bap_itercoeff_COEFF.h"
#include "bap_add_polynom_COEFF.h"
#include "bap_mul_polynom_COEFF.h"
#include "bap__check_COEFF.h"
#include "bap_geobucket_COEFF.h"

#define BAD_FLAG_COEFF

/*
 * $R = - A$.
 * If $A$ and $R$ are identical then the operation just consists in
 * changing the sign of the coefficients.
*/

/*
 * texinfo: bap_neg_polynom_COEFF
 * Assign @math{- A} to @var{R}.
 */

BAP_DLL void
bap_neg_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A)
{
  struct bap_itermon_COEFF iter;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF S;
  struct bav_term T;
  ba0_int_p nbmon;
  struct ba0_mark M;

  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (bap_is_zero_polynom_COEFF (A))
    bap_set_polynom_zero_COEFF (R);
  else if (R == A)
    {
      bap_begin_itermon_COEFF (&iter, A);
      while (!bap_outof_itermon_COEFF (&iter))
        {
          COEFF_neg (*bap_coeff_itermon_COEFF (&iter),
              *bap_coeff_itermon_COEFF (&iter));
          bap_next_itermon_COEFF (&iter);
        }
    }
  else
    {
      bap_begin_itermon_COEFF (&iter, A);

      nbmon = bap_nbmon_polynom_COEFF (A) - R->clot->alloc;
      nbmon = nbmon > 0 ? nbmon : 0;

      if (bap_are_disjoint_polynom_COEFF (R, A))
        {
          bap_begin_creator_COEFF (&crea, R, &A->total_rank,
              bap_exact_total_rank, nbmon);

          if (bap_is_write_allable_creator_COEFF (&crea, A))
            bap_write_neg_all_creator_COEFF (&crea, A);
          else
            {
              ba0_push_another_stack ();
              ba0_record (&M);
              bav_init_term (&T);
              bav_realloc_term (&T, A->total_rank.size);
              ba0_pull_stack ();

              while (!bap_outof_itermon_COEFF (&iter))
                {
                  bap_term_itermon_COEFF (&T, &iter);
                  bap_write_neg_creator_COEFF (&crea, &T,
                      *bap_coeff_itermon_COEFF (&iter));
                  bap_next_itermon_COEFF (&iter);
                }

              ba0_restore (&M);
            }
          bap_close_creator_COEFF (&crea);
        }
      else
        {
          ba0_push_another_stack ();
          ba0_record (&M);
          bav_init_term (&T);
          bav_realloc_term (&T, A->total_rank.size);

          bap_init_polynom_COEFF (&S);

          bap_begin_creator_COEFF (&crea, &S, &A->total_rank,
              bap_exact_total_rank, bap_nbmon_polynom_COEFF (A));
          while (!bap_outof_itermon_COEFF (&iter))
            {
              bap_term_itermon_COEFF (&T, &iter);
              bap_write_neg_creator_COEFF (&crea, &T,
                  *bap_coeff_itermon_COEFF (&iter));
              bap_next_itermon_COEFF (&iter);
            }

          bap_close_creator_COEFF (&crea);
          ba0_pull_stack ();

          bap_set_polynom_COEFF (R, &S);
          ba0_restore (&M);
        }
    }
}

/*
 * $R = c\,A$.
 * If $A$ and $R$ are identical then the operation just consists in multiplying
 * the coefficients by~$c$.
 */

/*
 * texinfo: bap_mul_polynom_numeric_COEFF
 * Assign @math{c\,A} to @var{R}.
 */

BAP_DLL void
bap_mul_polynom_numeric_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    COEFF_t c)
{
  struct bap_itermon_COEFF iter;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF S;
  struct bav_term T;
  COEFF_t prod, *lc;
  enum bap_typeof_total_rank type;
  ba0_int_p nbmon;
  struct ba0_mark M;

  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (COEFF_is_zero (c))
    bap_set_polynom_zero_COEFF (R);
  else if (COEFF_is_one (c))
    {
      if (R != A)
        bap_set_polynom_COEFF (R, A);
    }
  else if (R == A && ba0_domain_COEFF ())
    {
      bap_begin_itermon_COEFF (&iter, A);
      while (!bap_outof_itermon_COEFF (&iter))
        {
          lc = bap_coeff_itermon_COEFF (&iter);
          COEFF_mul (*lc, *lc, c);
          if (COEFF_is_zero (*lc))
            BA0_RAISE_EXCEPTION (BA0_ERRALG);
          bap_next_itermon_COEFF (&iter);
        }
    }
  else
    {
      bap_begin_itermon_COEFF (&iter, A);
      nbmon = bap_nbmon_polynom_COEFF (A) - R->clot->alloc;
      nbmon = nbmon > 0 ? nbmon : 0;

      type = ba0_domain_COEFF ()? bap_exact_total_rank : bap_approx_total_rank;

      if (bap_are_disjoint_polynom_COEFF (R, A))
        {
          bap_begin_creator_COEFF (&crea, R, &A->total_rank, type, nbmon);

          if (bap_is_write_allable_creator_COEFF (&crea, A)
              && ba0_domain_COEFF ())
            bap_write_mul_all_creator_COEFF (&crea, A, c);
          else
            {
              ba0_push_another_stack ();
              ba0_record (&M);

              bav_init_term (&T);
              bav_realloc_term (&T, A->total_rank.size);

              COEFF_init (prod);

              while (!bap_outof_itermon_COEFF (&iter))
                {
                  lc = bap_coeff_itermon_COEFF (&iter);
                  COEFF_mul (prod, c, *lc);
                  if (!COEFF_is_zero (prod))
                    {
                      bap_term_itermon_COEFF (&T, &iter);
                      ba0_pull_stack ();
                      bap_write_creator_COEFF (&crea, &T, prod);
                      ba0_push_another_stack ();
                    }
                  else if (ba0_domain_COEFF ())
                    BA0_RAISE_EXCEPTION (BA0_ERRALG);
                  bap_next_itermon_COEFF (&iter);
                }

              ba0_pull_stack ();
              ba0_restore (&M);
            }

          bap_close_creator_COEFF (&crea);
        }
      else
        {
          ba0_push_another_stack ();
          ba0_record (&M);

          bap_init_polynom_COEFF (&S);

          bav_init_term (&T);
          bav_realloc_term (&T, A->total_rank.size);

          COEFF_init (prod);

          bap_begin_creator_COEFF (&crea, &S, &A->total_rank, type,
              bap_nbmon_polynom_COEFF (A));

          while (!bap_outof_itermon_COEFF (&iter))
            {
              lc = bap_coeff_itermon_COEFF (&iter);
              COEFF_mul (prod, c, *lc);
              if (!COEFF_is_zero (prod))
                {
                  bap_term_itermon_COEFF (&T, &iter);
                  bap_write_creator_COEFF (&crea, &T, prod);
                }
              else if (ba0_domain_COEFF ())
                BA0_RAISE_EXCEPTION (BA0_ERRALG);
              bap_next_itermon_COEFF (&iter);
            }

          bap_close_creator_COEFF (&crea);

          ba0_pull_stack ();
          bap_set_polynom_COEFF (R, &S);
          ba0_restore (&M);
        }
    }
}

/* $R = A \, (x - \alpha)$ where $\mbox{\em val} = (x, \alpha)$.  */

/*
 * texinfo: bap_mul_polynom_value_int_p_COEFF
 * Assign @math{A \, (x - \alpha)} to @var{R}, where @math{(x,\, \alpha)}
 * denotes @var{val}.
 */

BAP_DLL void
bap_mul_polynom_value_int_p_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bav_value_int_p *val)
{
  struct bav_term T;
  struct bav_rank rg;
  COEFF_t c;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF *P;
  struct ba0_mark M;

  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);
  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&T);
  rg.var = val->var;
  rg.deg = 1;
  bav_set_term_rank (&T, &rg);
  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &T, bap_exact_total_rank, 2);
  COEFF_init_set_ui (c, 1);
  bap_write_creator_COEFF (&crea, &T, c);
  bav_set_term_one (&T);
  COEFF_set_si (c, val->value);
  bap_write_neg_creator_COEFF (&crea, &T, c);
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_mul_polynom_COEFF (R, A, P);
  ba0_restore (&M);
}

/* $R = A \, T$.  */

/*
 * texinfo: bap_mul_polynom_term_COEFF
 * Assign @math{A \, T} to @var{R}.
 */

BAP_DLL void
bap_mul_polynom_term_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bav_term *T)
{
  struct bap_itermon_COEFF iter;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF *P;
  struct bav_term U;
  struct ba0_mark M;

  bap__check_ordering_COEFF (A);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (bav_is_one_term (T))
    {
      if (R != A)
        bap_set_polynom_COEFF (R, A);
      return;
    }

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&U);
  bav_set_term (&U, &A->total_rank);
  bav_mul_term (&U, &U, T);
  bap_begin_itermon_COEFF (&iter, A);
  P = bap_new_polynom_COEFF ();
  bap_begin_creator_COEFF (&crea, P, &U, bap_exact_total_rank,
      bap_nbmon_polynom_COEFF (A));
  while (!bap_outof_itermon_COEFF (&iter))
    {
      bap_term_itermon_COEFF (&U, &iter);
      bav_mul_term (&U, &U, T);
      bap_write_creator_COEFF (&crea, &U, *bap_coeff_itermon_COEFF (&iter));
      bap_next_itermon_COEFF (&iter);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*
 * texinfo: bap_mul_polynom_variable_COEFF
 * Assign @math{A \, v^d} to @var{R}.
 */

BAP_DLL void
bap_mul_polynom_variable_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bav_variable *v,
    bav_Idegree d)
{
  struct bav_term term;
  struct ba0_mark M;

  ba0_push_another_stack ();
  ba0_record (&M);
  bav_init_term (&term);
  bav_set_term_variable (&term, v, d);
  ba0_pull_stack ();
  bap_mul_polynom_term_COEFF (R, A, &term);
  ba0_restore (&M);
}

/* $R = A \, c\,T$.  */

/*
 * texinfo: bap_mul_polynom_monom_COEFF
 * Assign @math{A \, c\, T} to @var{R}.
 */

BAP_DLL void
bap_mul_polynom_monom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    COEFF_t c,
    struct bav_term *T)
{
  struct bap_itermon_COEFF iter;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF *P;
  struct bav_term U;
  COEFF_t d;
  struct ba0_mark M;

  bap__check_ordering_COEFF (A);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (COEFF_is_zero (c))
    {
      bap_set_polynom_zero_COEFF (R);
      return;
    }
  else if (COEFF_is_one (c))
    {
      bap_mul_polynom_term_COEFF (R, A, T);
      return;
    }

  ba0_push_another_stack ();
  ba0_record (&M);

  bav_init_term (&U);
  bav_set_term (&U, &A->total_rank);
  bav_mul_term (&U, &U, T);
  bap_begin_itermon_COEFF (&iter, A);

  P = bap_new_polynom_COEFF ();
#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpq)
  bap_begin_creator_COEFF (&crea, P, &U, bap_exact_total_rank,
      bap_nbmon_polynom_COEFF (A));
#else
  bap_begin_creator_COEFF (&crea, P, &U, bap_approx_total_rank,
      bap_nbmon_polynom_COEFF (A));
#endif

  COEFF_init (d);

  while (!bap_outof_itermon_COEFF (&iter))
    {
      bap_term_itermon_COEFF (&U, &iter);
      bav_mul_term (&U, &U, T);
      COEFF_mul (d, c, *bap_coeff_itermon_COEFF (&iter));
      bap_write_creator_COEFF (&crea, &U, d);
      bap_next_itermon_COEFF (&iter);
    }
  bap_close_creator_COEFF (&crea);
  ba0_pull_stack ();
  bap_set_polynom_COEFF (R, P);
  ba0_restore (&M);
}

/*****************************************************************************
 MULTIPLICATION
 *****************************************************************************/

static void
bap_mul_elem_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bap_polynom_COEFF *B)
{
  struct bap_creator_COEFF crea;
  struct bap_itermon_COEFF iterA, iterB;
  struct bap_polynom_COEFF R1;
  struct bap_geobucket_COEFF geo;
  COEFF_t cz, *ca, *cb;
  struct bav_term TA, TB, TTB;
  enum bap_typeof_total_rank type;
  struct ba0_mark M;

  if (bap_nbmon_polynom_COEFF (A) > bap_nbmon_polynom_COEFF (B))
    BA0_SWAP (struct bap_polynom_COEFF *,
        A,
        B);

  type = ba0_domain_COEFF ()? bap_exact_total_rank : bap_approx_total_rank;

  ba0_push_another_stack ();
  ba0_record (&M);

  COEFF_init (cz);

  bav_init_term (&TTB);
  bav_set_term (&TTB, &B->total_rank);

  bav_init_term (&TA);
  bav_init_term (&TB);

  bap_init_geobucket_COEFF (&geo);
  bap_init_polynom_COEFF (&R1);

  bap_begin_itermon_COEFF (&iterA, A);
  while (!bap_outof_itermon_COEFF (&iterA))
    {
      ca = bap_coeff_itermon_COEFF (&iterA);
      bap_term_itermon_COEFF (&TA, &iterA);
      bav_mul_term (&TB, &TTB, &TA);
      bap_begin_creator_COEFF (&crea, &R1, &TB, type,
          bap_nbmon_polynom_COEFF (B));
      bap_begin_itermon_COEFF (&iterB, B);
      while (!bap_outof_itermon_COEFF (&iterB))
        {
          cb = bap_coeff_itermon_COEFF (&iterB);
          COEFF_mul (cz, *ca, *cb);
          bap_term_itermon_COEFF (&TB, &iterB);
          bav_mul_term (&TB, &TB, &TA);
          bap_write_creator_COEFF (&crea, &TB, cz);
          bap_next_itermon_COEFF (&iterB);
        }
      bap_close_creator_COEFF (&crea);
      bap_add_geobucket_COEFF (&geo, &R1);
      bap_next_itermon_COEFF (&iterA);
    }
  ba0_pull_stack ();
  bap_set_polynom_geobucket_COEFF (R, &geo);
  ba0_restore (&M);
}

/* $R = A \, B$.  */

/*
 * texinfo: bap_mul_polynom_COEFF
 * Assign @math{A\,B} to @var{R}.
 */

BAP_DLL void
bap_mul_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    struct bap_polynom_COEFF *B)
{
  struct bap_itercoeff_COEFF iterA, iterB;
  struct bap_itermon_COEFF iter;
  struct bap_creator_COEFF crea;
  struct bap_polynom_COEFF C, CA, CB;
  struct bap_polynom_COEFF *AA, *BB, *P;
  struct bav_term T, U, TA, TB;
  struct bav_variable *xa, *xb, *v;
  bav_Iordering r;
  ba0_int_p i;
  struct ba0_mark M;

  bap__check_compatible_COEFF (A, B);
  if (R->readonly)
    BA0_RAISE_EXCEPTION (BA0_ERRALG);

  if (bap_is_numeric_polynom_COEFF (B))
    BA0_SWAP (struct bap_polynom_COEFF *,
        A,
        B);
/*
   Si l'un des deux polynomes est constant, A est constant
*/
  if (bap_is_numeric_polynom_COEFF (A))
    {
      if (bap_is_zero_polynom_COEFF (A) || bap_is_zero_polynom_COEFF (B))
        bap_set_polynom_zero_COEFF (R);
      else
        {
          COEFF_t c;            /* Sinon bug quand A == R */
          ba0_push_another_stack ();
          ba0_record (&M);
          COEFF_init_set (c, *bap_numeric_initial_polynom_COEFF (A));
          ba0_pull_stack ();
          bap_mul_polynom_numeric_COEFF (R, B, c);
          ba0_restore (&M);
        }
      return;
    }

  if (bap_nbmon_polynom_COEFF (B) == 1)
    {
      BA0_SWAP (struct bap_polynom_COEFF *,
          A,
          B);
    }
/*
   Si l'un des deux polynomes est reduit a un monome alors A l'est.
   Optimisation mais non necessaire.
*/
  if (bap_nbmon_polynom_COEFF (A) == 1)
    {
      COEFF_t c;

      ba0_push_another_stack ();
      ba0_record (&M);
      bap_begin_itermon_COEFF (&iter, A);
      bav_init_term (&T);
      bap_term_itermon_COEFF (&T, &iter);
      COEFF_init_set (c, *bap_coeff_itermon_COEFF (&iter));
      ba0_pull_stack ();
      bap_mul_polynom_monom_COEFF (R, B, c, &T);
      ba0_restore (&M);
      return;
    }
/*
   Ni A ni B n'est constant
   On marque a 2 les variables communes aux deux polynomes et a 1 les autres
*/
  for (i = 0; i < A->total_rank.size; i++)
    A->total_rank.rg[i].var->hack = 0;
  for (i = 0; i < B->total_rank.size; i++)
    B->total_rank.rg[i].var->hack = 0;
  for (i = 0; i < A->total_rank.size; i++)
    A->total_rank.rg[i].var->hack++;
  for (i = 0; i < B->total_rank.size; i++)
    B->total_rank.rg[i].var->hack++;
/*
   Dans le bav_type_ordering r, les variables non communes sont plus grandes
	que les autres.
   xa = la plus petite struct bav_variable * de A non commune a B (ou nulle)
   xb = la plus petite struct bav_variable * de B non commune a A (ou nulle)
*/

  r = bav_R_copy_ordering (bav_R_Iordering ());
  bav_R_push_ordering (r);

  xa = BAV_NOT_A_VARIABLE;
  for (i = A->total_rank.size - 1; i >= 0; i--)
    {
      v = A->total_rank.rg[i].var;
      if (v->hack == 1)
        {
          if (xa == BAV_NOT_A_VARIABLE)
            xa = v;
          bav_R_set_maximal_variable (v);
        }
    }
  xb = BAV_NOT_A_VARIABLE;
  for (i = B->total_rank.size - 1; i >= 0; i--)
    {
      v = B->total_rank.rg[i].var;
      if (v->hack == 1)
        {
          if (xb == BAV_NOT_A_VARIABLE)
            xb = v;
          bav_R_set_maximal_variable (v);
        }
    }
/*
   Si l'un des deux polynomes a toutes ses variables dans l'autre alors
	c'est le cas pour B.
*/
  if (xa == BAV_NOT_A_VARIABLE)
    {
      BA0_SWAP (struct bap_polynom_COEFF *,
          A,
          B);
      BA0_SWAP (struct bav_variable *,
          xa,
          xb);
    }
/*
   Exit le cas de deux polynomes sur le meme alphabet
*/
  if (xa == BAV_NOT_A_VARIABLE)
    {
      bav_R_pull_ordering ();
      bav_R_free_ordering (r);
      bap_mul_elem_polynom_COEFF (R, A, B);
      return;
    }
/*
   A depend de variables non communes avec B
*/
  ba0_push_another_stack ();
  ba0_record (&M);

  {
    struct bap_polynom_COEFF *AA0 = bap_new_readonly_polynom_COEFF ();
    bap_sort_polynom_COEFF (AA0, A);
    AA = bap_new_polynom_COEFF ();
    bap_set_polynom_COEFF (AA, AA0);
  }

  bap_begin_itercoeff_COEFF (&iterA, AA, xa);

  if (xb != BAV_NOT_A_VARIABLE)
    {
      struct bap_polynom_COEFF *BB0 = bap_new_readonly_polynom_COEFF ();
      bap_sort_polynom_COEFF (BB0, B);
      BB = bap_new_polynom_COEFF ();
      bap_set_polynom_COEFF (BB, BB0);
    }
  else
    BB = B;

  bap_init_polynom_COEFF (&C);
  bap_init_polynom_COEFF (&CA);
  bap_init_polynom_COEFF (&CB);
  bav_init_term (&T);
  bav_init_term (&TA);
  bav_init_term (&TB);
  bav_init_term (&U);

  bav_mul_term (&T, &AA->total_rank, &BB->total_rank);
  i = BA0_MAX (bap_nbmon_polynom_COEFF (AA), bap_nbmon_polynom_COEFF (BB));
/*
 * Note: the monomials of P will temporarily not be sorted
 */
  P = bap_new_polynom_COEFF ();
#if defined (BAD_FLAG_mpz) || defined (BAD_FLAG_mpq)
  bap_begin_creator_COEFF (&crea, P, &T, bap_exact_total_rank, i);
#else
  bap_begin_creator_COEFF (&crea, P, &T, bap_approx_total_rank, i);
#endif

  while (!bap_outof_itercoeff_COEFF (&iterA))
    {
      bap_coeff_itercoeff_COEFF (&CA, &iterA);
      bap_term_itercoeff_COEFF (&TA, &iterA);
      bap_begin_itercoeff_COEFF (&iterB, BB,
          xb == BAV_NOT_A_VARIABLE ? xa : xb);
      while (!bap_outof_itercoeff_COEFF (&iterB))
        {
          bap_coeff_itercoeff_COEFF (&CB, &iterB);
          bap_term_itercoeff_COEFF (&TB, &iterB);

          bav_mul_term (&T, &TA, &TB);
          bap_mul_elem_polynom_COEFF (&C, &CA, &CB);

          bap_begin_itermon_COEFF (&iter, &C);
          while (!bap_outof_itermon_COEFF (&iter))
            {
              COEFF_t *c;

              c = bap_coeff_itermon_COEFF (&iter);
              bap_term_itermon_COEFF (&U, &iter);
              bav_mul_term (&U, &U, &T);
              bap_write_creator_COEFF (&crea, &U, *c);
              bap_next_itermon_COEFF (&iter);
            }
          bap_next_itercoeff_COEFF (&iterB);
        }
      bap_next_itercoeff_COEFF (&iterA);
    }
  bap_close_creator_COEFF (&crea);
  bav_R_pull_ordering ();
/*
 * Now, the monomials get sorted
 */
  bap_physort_polynom_COEFF (P);

  bav_R_free_ordering (r);
  ba0_pull_stack ();

  i = BA0_MAX (bap_nbmon_polynom_COEFF (A), bap_nbmon_polynom_COEFF (B));

  bap_set_polynom_COEFF (R, P);

  ba0_restore (&M);
}

/* $R = A^n$.  */

/*
 * texinfo: bap_pow_polynom_COEFF
 * Assign @math{A^d} to @var{R}.
 */

BAP_DLL void
bap_pow_polynom_COEFF (
    struct bap_polynom_COEFF *R,
    struct bap_polynom_COEFF *A,
    bav_Idegree n)
{
  struct bap_polynom_COEFF E, F;
  bav_Idegree p;
  bool E_vaut_un;
  struct ba0_mark M;

  if (n == 0)
    bap_set_polynom_one_COEFF (R);
  else if (n == 1)
    {
      if (R != A)
        bap_set_polynom_COEFF (R, A);
    }
  else
    {
      ba0_push_another_stack ();
      ba0_record (&M);

      if (n % 2 == 1)
        {
          bap_init_polynom_COEFF (&E);
          bap_set_polynom_COEFF (&E, A);
          E_vaut_un = false;
        }
      else
        E_vaut_un = true;

      bap_init_polynom_COEFF (&F);
      bap_mul_polynom_COEFF (&F, A, A);

      for (p = n / 2; p != 1; p /= 2)
        {
          if (p % 2 == 1)
            {
              if (E_vaut_un)
                {
                  bap_init_polynom_COEFF (&E);
                  bap_set_polynom_COEFF (&E, &F);
                  E_vaut_un = false;
                }
              else
                bap_mul_polynom_COEFF (&E, &F, &E);
            }
          bap_mul_polynom_COEFF (&F, &F, &F);
        }
      ba0_pull_stack ();
      if (E_vaut_un)
        bap_set_polynom_COEFF (R, &F);
      else
        bap_mul_polynom_COEFF (R, &E, &F);
      ba0_restore (&M);
    }
}

#undef BAD_FLAG_COEFF
