#include "bap.h"

typedef void ftype (
    struct bap_polynom_mpz *,
    struct bap_polynom_mpz *);

static struct
{
  char *ident;
  ftype *f;
} T[] = {
  {"initial", &bap_initial_polynom_mpz},
  {"separant", &bap_separant_polynom_mpz},
  {"reductum", &bap_reductum_polynom_mpz},
  {"sepuctum", &bap_sepuctum_polynom_mpz},
  {(char *) 0, 0}
};

int
main (
    int argc,
    char **argv)
{
  struct bap_tableof_polynom_mpz succeed, fail, result, nulles;
  bav_Iordering r;
  struct ba0_table blks;
  struct ba0_tableof_string ders, pars;
  FILE *f;
  volatile bool b = true;
  char nota[64], dring[64], singsol[64];
  static char buffer[32768], expected[32768];
  char ident[64];
  ba0_int_p i;
  ftype *func;

/*
 * Are we able to open the file?
 */
  if (argc != 2)
    {
      fprintf (stderr, "error: missing filename on command line\n");
      return 1;
    }
  f = fopen (argv[1], "r");
  if (f == NULL)
    {
      fprintf (stderr, "unable to open: %s\n", argv[1]);
      return 1;
    }

  bap_restart (0, 0);

  BA0_TRY
  {
    ba0_set_input_FILE (f);
/*
 * R = DifferentialRing (...)
 */
    ba0_init_table (&blks);
    ba0_init_table ((struct ba0_table *) &pars);
    ba0_init_table ((struct ba0_table *) &ders);

    ba0_get_token_analex ();
    ba0_scanf
        ("R = DifferentialRing (blocks = %t[%b], parameters = %t[%six], derivations = %t[%s], notation = %s)",
        &blks, &pars, &ders, nota);
/*
 * Do not bother with nota
 */
    bav_set_settings_variable (&bav_scanf_maple_all_variable, 0, 0, 0, 0);
    bav_set_settings_rank (&bav_printf_stars_rank);
/*
 * The ordering
 */
    ba0_sprintf (buffer,
        "ordering (derivations = %t[%s], blocks = %t[%b])\n", &ders, &blks);
    ba0_scanf_printf ("%ordering",
        "ordering (derivations = %t[%s], blocks = %t[%b])", &r, &ders, &blks);
    bav_R_push_ordering (r);
/*
 * The parameters
 */
    ba0_sprintf (buffer, "parameters = %t[%s]\n", &pars);
    ba0_scanf_printf ("%t[%param]", "%t[%s]", &bav_global.parameters, &pars);
/*
 * The lists succeed, fail, result
 */
    ba0_init_table ((struct ba0_table *) &succeed);
    ba0_get_token_analex ();
    ba0_scanf ("succeed = %t[%Az]", &succeed);

    ba0_init_table ((struct ba0_table *) &result);
    ba0_get_token_analex ();
    ba0_scanf ("result = %t[%Az]", &result);

    ba0_init_table ((struct ba0_table *) &fail);
    ba0_get_token_analex ();
    ba0_scanf ("fail = %t[%Az]", &fail);

    ba0_init_table ((struct ba0_table *) &nulles);
    bap_zero_derivatives_of_tableof_parameter_mpz (&nulles,
        &bav_global.parameters);
/*
 * The command
 */
    ba0_get_token_analex ();
    ba0_scanf ("function = %s", ident);
    fclose (f);

    i = 0;
    while (T[i].ident && strcmp (T[i].ident, ident) != 0)
      i += 1;
    if (T[i].ident == (char *) 0)
      BA0_RAISE_EXCEPTION (BA0_ERRALG);
    func = T[i].f;

    if (succeed.size != result.size)
      BA0_RAISE_EXCEPTION (BA0_ERRALG);

    b = true;
    for (i = 0; b && i < succeed.size; i++)
      {
        struct bap_polynom_mpz R;
        bap_init_polynom_mpz (&R);
        (*func) (&R, succeed.tab[i]);
        b = bap_equal_polynom_mpz (&R, result.tab[i]);
      }
    for (i = 0; b && i < fail.size; i++)
      {
        struct bap_polynom_mpz R;
        bap_init_polynom_mpz (&R);
        BA0_TRY
        {
          (*func) (&R, fail.tab[i]);
          b = false;
        }
        BA0_ENDTRY;
      }
  }
  BA0_CATCH
  {
    fprintf (stderr, "syntax error in the test file\n");
    b = false;
  }
  BA0_ENDTRY;

  bap_terminate (ba0_init_level);
  return b ? 0 : 1;
}
