The BMI library (BLAD Multiple Interfaces) is an interface library between 
various scientific computing software and the BLAD libraries.

The BMI library contains C code as well as Maple / Cython code

Building from the development version
-------------------------------------

The BMI library has an autoconf/automake/libtool based configuration system.

However, if you download the development version from the git repository,
the configure file is missing. In order to create it, you need the
autotools. Then go to the BMI top directory and run

../blad/tools/autogen

Building the library (Python/sympy version)
-------------------------------------------

Before building BMI, you need to build BLAD.

It is advised to create a separate bmi-build directory in the
DifferentialAlgebra git working copy. At some point, the directory
tree could look like this:

DifferentialAlgebra
├── blad
├── blad-build
├── blad-install
├── bmi
├── bmi-build-sympy
├── bmi-install-sympy
├── gallery
├── LICENSE
└── README.md

Assume the current directory is DifferentialAlgebra.
On a Linux machine, run

mkdir bmi-build-sympy
cd bmi-build-sympy
../bmi/tools/bmi_configure_for_sympy
make
make pdf
make check
make install

The make check command runs a development version of the bmi library.
The make install command installs BMI in bmi-install.

In order to get the sympy package in the Python environment:

cd ../bmi-install-sympy/DifferentialAlgebra
python3 setup.py build
pip install . --upgrade

Testing
-------

The following commands should then work:

$ python3
>>> from sympy import *
>>> from DifferentialAlgebra import *
>>> x = var('x')
>>> y = indexedbase('y')
>>> R = DifferentialRing (derivations = [x], blocks = [y], notation = 'jet')
>>> ideal = R.RosenfeldGroebner([y[x]**2 - 4*y])
>>> [ C.equations (solved=True) for C in ideal ]
[[Eq(y[x]**2, 4*y)], [Eq(y, 0)]]

