# Local to DifferentialRing

# This function returns a sequence of two lists: freezeseq, unnumberseq
# 
# Both lists are meant to be used by subs, in order to perform some renaming.
# freezeseq freezes some identifiers which are not parsed by the BLAD parsers.
# It is used by ConvertString.
# 
# The BLAD functions return symbols under numbered form.
# unnumberseq permits to convert them back to their values.
# It us used by BMIBladEvalSafe.

FrozenSymbols := proc (derivations, blocks)
    local symbols, arguments, frozen_indices, freezeseq, unnumberseq, i;

    symbols := map (proc (block) 
			if IsCompositeBlock (block) then
			    return op (block);
			else
			    return block;
			end if
                    end proc, blocks);
    symbols := [ op (derivations), op (symbols) ];
    if nops (convert (symbols, ':-set')) <> nops (symbols) then
	error _ERRDUPS_;
    end if;
    arguments := "FrozenSymbols" (op (map (ToString, symbols)));
    arguments := arguments,
            ToBMIOptionSequence (':-_JET_', ':-_JET_', 0, 0, ':-_SMALL_');
# Do not call BMIBladEvalSafe !
    frozen_indices := BMIBladEval (arguments);

    freezeseq := seq (symbols [frozen_indices [i]] = 
			cat (`__`, frozen_indices [i]-1),
		 i = 1 .. nops (frozen_indices));
    unnumberseq := seq (cat (`_`, i-1) = symbols [i], i = 1 .. nops (symbols));
    return [ freezeseq ], [ unnumberseq ];
end proc:

# Local to FrozenSymbols
# This task should be performed at the bmi level, in the future.

IsCompositeBlock := proc (block)
    if type (block, ':-list') then
	return true;
    elif not type (block, ':-indexed') then
	return false;
    else
	return member (op (0, block), [':-grlexA', ':-grlexB', 
				':-degrevlexA', ':-degrevlexB', ':-lex']);
    end if;
end proc: 
