##TOPIC(help) DifferentialAlgebra0[Glossary]
##SECTION Glossary
##
##- This dictionary provides definitions for some words or expressions,
##  commonly used in the package documentation. Some of these words 
##  (e.g. *order*, *derivative*), are used, sometimes in their technical 
##  sense, sometimes in their common sense. 
##  These words or expressions are *emphasized* when one wants to stress 
##  the fact that the technical sense is meant.
##
##-- **attribute**. In this package, an *attribute* is a property of a regular 
##  differential chain.  Seven attributes are defined: 
##  *differential*, *prime*, *primitive*, *squarefree*, *coherent*,
##  *autoreduced* and *normalized*. 
##  The two first attributes provide properties of the ideal defined 
##  by the regular differential chain. 
##  The five other ones provide properties of the differential polynomials 
##  which constitute the chain. 
##
##-- **autoreduced**. An *attribute* of regular differential chains.
##  It indicates that the regular differential chain is autoreduced, in
##  the sense, that, for all differential polynomials _p_ and _q_ of the 
##  chain, the degree of _p_ in the leading derivative of _q_ is less than 
##  the degree of _q_ in its leading derivative.
##
##-- **BLAD**. Open source libraries, written in the C programming
##  language, dedicated to differential elimination. The 
##  `DifferentialAlgebra0` package relies on the `BLAD` libraries
##  for most computations.
##
##-- **block**. A list of dependent variables plus a *block-keyword*.
##  Blocks appear in the definition of *rankings*.
##
##-- **block-keyword**. A keyword, which precises the *ranking* of 
##  a *block*. Five *block-keywords* are defined: `grlexA`, `grlexB`,
##  `degrevlexA`, `degrevlexB` and `lex`. 
##
##-- **characteristic set**. A particular case of a regular differential chain.
##
##-- **coherent**. An *attribute* of regular differential chains.
##  It indicates that the regular differential chain is coherent.
##  This notion is detailed in a following section.
##
##-- **component**. In an intersection of differential ideals, presented
##  by regular differential chains, one of the components of the intersection.
##  
##-- **constant**. An expression whose derivatives are all equal to zero.
##
##-- **degrevlexA**. One of the *block-keywords*.
##
##-- **degrevlexB**. One of the *block-keywords*.
## 
##-- **Delta-polynomial**. The _Delta_-polynomial defined by two 
##  differential polynomials is important for testing the *coherence* 
##  property of systems of partial differential equations. 
##  Somehow, it plays the same role as the _S_-polynomials of the
##  Groebner bases theory.
##  Let _A[i]_ and _A[j]_ be two differential polynomials, whose leading 
##  derivatives _v[i]_ and _v[j]_ are
##  derivatives of some same dependent variable _u_. 
##  Denote _theta[i]_ and _theta[j]_ the derivation operators associated
##  to _v[i]_ and _v[j]_ and _theta[i,j]_ their least common multiple.
##
##--- If _theta[i,j]_ is different from both _theta[i]_ and _theta[j]_,
##   then the _Delta_-polynomial defined by
##   _A[i]_ and _A[j]_ is the differential polynomial
##   _S[i]_ _theta[i,j] / theta [j]_ _A[j]_ -
##   _S[j]_ _theta[i,j] / theta [i]_ _A[i]_ where _S[i]_ and _S[j]_
##   denote the separants of _A[i]_ and _A[j]_ and the left multiplication
##   by a derivation operator stands for a differentiation.
##  
##--- If _theta[i,j]_ is equal to (say) _theta[j]_ then the _Delta_-polynomial 
##   defined by _A[i]_ and _A[j]_ is the pseudo-remainder of
##   _A[j]_ by _theta[j] / theta[i]_ _A[i]_ with respect to _v[j]_.
##   
##-- **dependent variable**. A variable which depends on the independent
##  variables, i.e., a function of the independent variables. In 
##  the classical books of *differential algebra*, it would be called 
##  a *differential indeterminate*.
##
##-- **dependent variable associated to a derivative**.
##  If _v_ is a derivative of a dependent variable _u_, then, _u_ is
##  said to be the dependent variable associated to _v_.
##
##-- **derivation**. In this package, derivations are taken with respect
##  to independent variables and are supposed to commute. 
##  The set of the derivations generates the monoid (semigroup) of the
##  derivation operators, which are denoted multiplicatively.
##  In the classical books of *differential algebra*, derivations are 
##  just abstract operations which satisfy the axioms of derivations.
##
##-- **derivation operator**. A power product of independent variables
##  denoting iterated derivations. Example: differentiating a differential
##  polynomial _p_, with respect to the derivation operator _x^2*y_,
##  consists in differentiating _p_ twice with respect to _x_, then 
##  once with respect to _y_. The identity derivation operator
##  is denoted _1_.
##
##-- **derivation operator associated to a derivative**.
##  If _v_ is a derivative of a dependent variable _u_, then, there
##  exists some derivation operator _theta_, such that differentiating
##  _u_ with respect to _theta_, gives _v_. The derivation operator _theta_
##  is called the derivation operator associated to _v_.
##  
##-- **derivative**. A derivative of a dependent variable. 
##
##-- **differential**. One of the *attributes* of regular differential
##  chains. It indicates that the ideal defined by the chain is
##  differential. The presence of *differential* implies the
##  presence of *squarefree*, and that of *coherent* in the
##  partial differential case.
##
##-- **differential algebra**. The mathematical theory that provides the
##  theoretical basis for this package. 
##
##-- **differential ideal**. Mathematically, an ideal which contains
##  the derivatives of some element _p_, whenever it contains _p_. 
##  In this package,
##  a differential ideal is a polynomial ideal which is presented
##  either by a single regular differential chain or by a list
##  of regular differential chains. Lists of regular differential chains
##  represent the intersection of the differential ideals defined by the 
##  chains. These ones are called the *components* of the intersection.
##  The chains which occur in a list must belong to the same differential 
##  polynomial ring. The empty list denotes the unit differential ideal.
##  The zero differential ideal can be represented by a regular differential
##  chain. The notion of differential ideal is detailed in a following section.
##
##-- **differential indeterminate**. An abstract symbol standing for a function,
##  over which the derivations act. In this package, the expression
##  *dependent variable* is preferred.
##
##-- **differential polynomial**. A polynomial, with (usually) rational 
##  coefficients, whose variables are either *derivatives* or
##  independent variables. In some contexts, the field of coefficients
##  may be a non-trivial differential field.
##
##-- **differential ring**. Mathematically, a ring endowed with finitely
##  many derivations, which (in our case) are supposed to commute. In this package,
##  a differential ring is a data structure representing a polynomial
##  differential ring, endowed with a ranking, and other minor features.
##  In the case of only one derivation, the ring is said to be *ordinary*.
##  In the case of two or more derivations, it is said to be *partial*.
##  In this package, a ring may be endowed with no derivation. In this case,
##  it is no more differential, and, the use of other packages, such as
##  "RegularChains" or "Groebner" is recommended.
##
##-- **grlexA**. One of the *block-keywords*.
##
##-- **grlexB**. One of the *block-keywords*.
## 
##-- **inconsistent**. An inconsistent system of differential polynomials
##  is a system which has no solution at all. The differential ideal
##  generated by an inconsistent system is the unit ideal, which is
##  presented, in this package, by the empty list.
##
##-- **independent variable**. A variable, with respect to which derivations
##  are taken. Unless stated otherwise, a dependent variable is supposed 
##  to depend on all the independent variables. Common examples are
##  the time _t_, and the space variables _x_, _y_, _z_.
##
##-- **initial**. The initial of a non-numeric differential polynomial _p_ is 
##  the leading coefficient of _p_, regarded as a univariate polynomial
##  with respect to its leading derivative.
##
##-- **irredundant**. A representation of an ideal _J_ as an intersection
##  of ideals _I[1]_, `...`, _I[n]_ is said to be irredundant if,
##  given any two different indices _j,k_
##  the ideal _I[j]_ is not included in the ideal _I[k]_.
##  In general, the representations computed by the "RosenfeldGroebner"
##  algorithm *are* redundant. This issue is addressed in a following section.
##
##-- **leading derivative**. The leading derivative _v_ of a non-numeric
##  differential polynomial _p_, is the highest *derivative*, with respect 
##  to some given ranking, such that the degree of _p_ in _v_ is positive.
##  In this package, the leading derivative of differential 
##  polynomials which only depend on independent variables is defined. 
##
##-- **leading rank**. The leading rank of a differential polynomial _p_
##  is the *rank* _v^d_ such that _v_ is the leading derivative of _p_,
##  and, _d_ is the degree of _p_ in _v_. 
##  In this package, the leading rank of differential polynomials which do
##  not depend on any derivative is defined. In particular, the leading 
##  rank of _0_ is _0_, the one of any other rational number is _1_.
##
##-- **lex**. One of the *block-keywords*.
##
##-- **normalized**. An *attribute* of regular differential chains.
##  It indicates that the regular differential chain is normalized,
##  in the sense that, the initials of the differential polynomials
##  of the chain do not depend on any leading derivative of any
##  element of the chain. 
##  The presence of *normalized* implies the presence of *autoreduced*
##  and *primitive*. 
##  Regular differential chains which hold these three *attributes* are 
##  canonical representatives of the ideals that they define (in the sense
##  that they only depend on the ideal and on the ranking).
##
##-- **numeric**. A numeric differential polynomial is a differential
##  polynomial which does not depend on any *derivative* or any independent 
##  variable.
##
##-- **order**. The order of a *derivative* is the total degree of
##  its associated derivation operator. The order of a differential
##  polynomial is the maximum of the orders of the *derivatives*
##  it depends on.
##
##-- **orderly**. A ranking is said to be orderly if, for all *derivatives* 
##  _u_ and _v_, _u_ is higher than _v_ whenever the *order*  of
##  _u_ is greater than the order of _v_.
##
##-- **ordinary**. An ordinary differential ring is a ring endowed
##  with a single derivation.
##
##-- **partial**. A partial differential ring is a ring endowed
##  with two or more derivations.
##
##-- **prime**. An ideal is said to be prime if, whenever it involves
##  some product _p*q_, it involves at least one the factors.
##  Any prime ideal is radical. A prime differential ideal is 
##  a differential ideal, which is prime.
##
##-- **prime**. One of the *attributes* of regular differential chains.
##  It indicates that the ideal defined by the chain is prime.
##  The presence of *prime* implies the presence of *squarefree*.
##  Some functions of the package, such as, "Pardi", only apply to
##  prime ideals. Many computations on regular differential chains
##  are simplified, when the ideal that they define are known to
##  be prime.
##
##-- **primitive**. An attribute of regular differential chains.
##  It indicates that each differential polynomial of the chain
##  is primitive, in the sense that the gcd of its coefficients 
##  is equal to _1_. 
##  Chain differential polynomials are regarded as univariate 
##  polynomials in their leading derivatives. Their coefficients
##  are viewed as multivariate polynomials over the ring of the integers.
##
##-- **proper derivative**. A *derivative* _v_ is said to be a *proper*
##  derivative of a *derivative* _u_, if _v_ is a derivative of _u_
##  and is different from _u_.
##
##-- **radical**. An ideal _I_ of a ring _R_ is said to be radical if 
##  it involves some element _p_ whenever it involves any power _p^d_ 
##  of _p_, where _d_ is any nonnegative integer. 
##  The radical of an ideal _I_ is the set of all the elements _p_ of _R_,
##  a power of which belongs to _I_. The radical of an ideal is an ideal.
##  The radical of a differential ideal is a differential ideal.
##
##-- **rank**. A derivative, or, an independent variable, raised to some 
##  positive integer. In addition, the two special 
##  ranks _0_ and _1_ are defined. Any ranking extends to a total
##  ordering on ranks: _0_ is less than _1_, which is less than any other rank.
##  Two ranks _u^n_ and _v^m_ are compared by comparing, first, _u_ and _v_,
##  then, the exponents _n_ and _m_.
##
##-- **ranking**. Any total ordering over the set of the *derivatives*, which
##  satisfies the two axioms of rankings. In this package, rankings are 
##  extended to the set of the independent variables. 
##  This notion is detailed in a following section.
##
##-- **redundant**. Not irredundant. See this word.
##
##-- **redundant component**. In an intersection of differential ideal,
##  a *component* which contains another *component*.
##
##-- **regular differential chain**. 
##  A data structure containing a list of differential polynomials 
##  sorted by increasing rank, plus some minor features. 
##  A few variants of regular differential chain are implemented. 
##  These variants can be selected by customizing the *attributes*
##  of the chain.
##  This notion is detailed in a following section.
##
##-- **saturation**. If _I_ is an ideal of a ring _R_ and _M_ is a 
##  multiplicative family of _R_, then, the saturation of _I_ by _M_ is 
##  the set _J_ of all the elements _p_ of _R_ such that, _m*p_ belongs
##  to _I_, for some _m_ of _M_. The saturation of an ideal is an ideal.
##  The saturation of a differential ideal is a differential ideal.
##
##-- **separant**. The separant of a non-numeric differential polynomial
##  _p_, is the partial derivative of _p_ with respect to its leading
##  derivative.
##
##-- **squarefree**. An *attribute* of regular differential chains.
##  It indicates that the regular differential chain is squarefree.
##  This notion is detailed in a following section.
##  
##-- **tail**. The tail of a non-numeric differential polynomial
##  _p_, is equal to _p - c*v^d_ where _c_ is the initial of _p_
##  and _v^d_ is its leading rank. 
##
##SECTION Complements
##
##- This section develops some notions mentioned in the dictionary.
##
##SUBSECTION Rankings
##
##- A *ranking* is any total ordering over the set of 
##  the *derivatives* (in this package, rankings are extended to the 
##  independent variables), which satisfies the two axioms of rankings:
##
##-- Every *derivative* _u_ is less than any of its proper derivatives.
##
##-- If _u_ and _v_ are derivatives such that _u_ is less than _v_ and _x_ is
##   any independent variable then the derivative of _u_ with respect to _x_
##   is less than the derivative of _v_ with respect to _x_.
##
##- In this package, rankings are defined by 
##  the list _x[1]_ `>` `...` `>` _x[p]_ of the independent variables plus 
##  a list _b[1]_ `>>` `...` `>>` _b[n]_ of *blocks*. 
##  Each *block* _b_ is defined by a list _u[1]_ `>` `...` `>` _u[m]_
##  of dependent variables plus a *block-keyword*, which is 
##  `grlexA`, `grlexB`, `degrevlexA`, `degrevlexB` or `lex`.
##  Any dependent variable must appear in exactly one block.
##
##- The `>>` operator between blocks indicates a *block elimination*
##  ranking: if _b[i]_ `>>` _b[j]_ are two blocks, _v[i]_ is any
##  derivative of any dependent variable occuring in _b[i]_, and,
##  _v[j]_ is any derivative of any dependent variable occuring in _b[j]_,
##  then _v[i]_ `>` _v[j]_.
##
##- Within a given block _b_ = _u[1]_ `>` `...` `>` _u[m]_, in
##  the ordinary differential case (only one derivation), the
##  *derivatives* are ordered by the unique *orderly*
##  ranking such that _u[1]_ `>` `...` `>` _u[m]_.
##
##- Within a given block _b_ = _u[1]_ `>` `...` `>` _u[m]_, in the
##  partial differential case (two or more derivations), the
##  *block-keyword* of _b_ is necessary to precise the ranking.
##  Consider two derivatives _v[i]_ and _v[j]_ of two
##  dependent variables _u[i]_ and _u[j]_, such that _i_ `>=` _j_.
##  Denote _theta[i]_ and _theta[j]_ the derivation operators
##  associated to these two derivatives. 
##  In the sequel, _theta[i] > theta[j]_ (lex) means that 
##  _theta[i] > theta[j]_ with respect to the lexicographical order of 
##  the "Groebner" bases theory, defined by _x[1]_ `>` `...` `>` _x[p]_,
##  while, _theta[i] > theta[j]_ (degrevlex) means that _theta[i] > theta[j]_ 
##  with respect to the degree reverse lexicographic order.
##  Depending on the *block-keyword* of _b_, listed below,
##  one has _v[i] > v[j]_ if the following condition is satisfied:
##
##-- **grlexA**.  
##  If the *order* of _v[i]_ is greater than the one of _v[j]_ else,
##  if the *orders* are equal and _i < j_ else,
##  if the *orders* are equal and _i = j_ and _theta[i] > theta[j]_ (lex).
##
##-- **grlexB**.
##  If the *order* of _v[i]_ is greater than the one of _v[j]_ else,
##  if the *orders* are equal and _theta[i] > theta[j]_ (lex) else,
##  if _theta[i] = theta[j]_ and _i < j_.
##
##-- **degrevlexA**.
##  If the *order* of _v[i]_ is greater than the one of _v[j]_ else,
##  if the *orders* are equal and _i < j_ else,
##  if the *orders* are equal and _i = j_ and _theta[i] > theta[j]_ (degrevlex).
##
##-- **degrevlexB**.
##  If the *order* of _v[i]_ is greater than the one of _v[j]_ else,
##  if the *orders* are equal and _theta[i] > theta[j]_ (degrevlex) else,
##  if _theta[i] = theta[j]_ and _i < j_.
##
##-- **lex**.
##  If _theta[i] > theta[j]_ (lex) else, 
##  if _theta[i] = theta[j]_ and _i < j_.
##
##ENDSUBSECTION
##
##SUBSECTION Regular differential chains
##
##- Regular differential chains generalize the regular chains of the
##  non-differential polynomial algebras (see the "RegularChains" package
##  and [ALM99])
##  and the characteristic sets of the classical differential algebra. 
##  In the sequel, one assumes that a *ranking* is fixed,
##  and, therefore, that each non-numeric differential polynomial
##  admits a leading derivative.
##  
##- A regular differential chain appears as a list of 
##  differential polynomials _A[1]_, `...`, _A[r]_ with rational numbers
##  for coefficients. 
##  Each differential polynomial _A[i]_ admits a leading derivative, _v[i]_
##  which is a *derivative*. The list _A[1]_, `...`, _A[r]_
##  is sorted by increasing leading rank. 
##  
##- The set _A[1]_, `...`, _A[r]_ is **differentially triangular**
##  and **partially autoreduced**:
##-- Given any two different indices _i,j_, the leading derivative _v[i]_
##   is not a derivative of _v[j]_.
##-- Given any two indices _i,j_, the differential polynomial _A[i]_
##   does not depend on any proper derivative of _v[i]_.
##  
##- At this stage, one needs to introduce the polynomial ring 
##  _R[0]_ = _k_ [_v[1]_, `...`, _v[r]_], which is obtained
##  by moving into the base field, all the independent variables
##  and derivatives, the _A[i]_ depend on, but which are not the
##  leading derivative of any _A[i]_.
##  
##- The set _A[1]_, `...`, _A[r]_ is a **squarefree regular chain**:
##-- Given any index _i_, the initial of _A[i]_ is an invertible element 
##   of _R[0]_/_I_ where _I_ denotes the ideal generated by 
##   _A[1]_, `...`, _A[i-1]_ in the ring _R[0]_. 
##-- Given any index _i_, the separant of _A[i]_
##   is an invertible element of _R[0]_/_I_ where _I_ denotes the 
##   ideal generated by _A[1]_, `...`, _A[i]_ in the ring _R[0]_.
##  
##- At this stage, one needs to introduce the polynomial system _A[v]_
##  where _v_ is a *derivative*. It is the set of all the derivatives
##  of the differential polynomials _A[i]_ whose leading derivatives
##  are strictly less than _v_. One also introduces the ring _R[v]_
##  obtained by moving into the base field of the polynomials, all
##  the independent variables and the derivatives, the elements of _A[v]_
##  depend on, but which are not the leading derivative of any element of _A[v]_.
##  
##- In the case of partial differential systems, the set _A[1]_, `...`, _A[r]_ 
##  is **coherent**: given any two different indices _i,j_, such that
##  _v[i]_ and _v[j]_ are derivatives of some same dependent variable _u_,
##  the _Delta_-polynomial defined by _A[i]_ and _A[j]_ is zero in
##  the ring _R[theta[i,j]*u]_/_I_ where _theta[i,j]*u_ is least common
##  derivative of _v[i]_ and _v[j]_ and _I_ denotes the ideal generated
##  by _A[theta[i,j]*u]_ in _R[theta[i,j]*u]_.
##
##ENDSUBSECTION
##
##SUBSECTION Differential ideals
##
##- Every regular differential chain _A[1]_, `...`, _A[r]_, of a 
##  polynomial differential ring _R_, defines a differential ideal _I_, 
##  which is the set of all the differential polynomials _F_
##  such that, for some power product _H_ of the initials and the 
##  separants of the _A[i]_, the differential polynomial _H*F_ is a finite 
##  linear combination of the derivatives
##  of the _A[i]_, with differential polynomials of _R_ for coefficients.
##
##- One stresses the fact the chain _A[1]_, `...`, _A[r]_ is *not*
##  a generating family of _I_. However, it completely defines _I_
##  and permits to compute a normal form of the residue class
##  of any differential polynomial in _R_/_I_. See "NormalForm". In particular:
##
##-- It permits to decide membership in _I_, i.e., zero in _R_/_I_:
##  a differential polynomial _F_ belongs to _I_ if and only if its 
##  normal form is _0_.
##
##-- It permits to decide if a differential polynomial _F_ is
##  regular modulo _I_, i.e., a non-zero divisor in _R_/_I_.
##
##- Given any system of differential polynomials _F[1]_, `...`, _F[s]_
##  and any ranking, it is possible to compute a representation of
##  the radical _J_ of the differential ideal generated by this system,
##  as a finite intersection of radical differential ideals 
##  _I[1]_, `...`, _I[n]_, presented by regular differential 
##  chains _C[1]_, `...`, _C[n]_. See "RosenfeldGroebner" [BLOP95,BLOP09].
##  See also [W98,LW99,H00,BKM01].
##  This representation permits to decide membership in _J_: a 
##  differential polynomial _F_ belongs to _J_ if and only if it
##  belongs to each differential ideal _I[k]_, i.e., if and only if its 
##  normal forms, with respect to all the regular differential chains 
##  _C[k]_, are all _0_. See "BelongsTo".
## 
##- One stresses the fact that, in general, the computed representation
##  _C[1]_, `...`, _C[n]_ is *redundant*. Indeed, the problem of
##  deciding whether two differential ideals presented by regular
##  differential chains are included in each other, is still open.
##  In the particular case of a differential ideal generated by a 
##  single differential polynomial, this problem is, however, solved,
##  thanks to the Low Power Theorem. See "EssentialComponents" [H99].
##
##- During the decomposition process, in general, the coefficients of the 
##  differential polynomials are assumed to lie in the 
##  field _Q_ (_x[1]_, `...`, _x[p]_) obtained by adjoining
##  the independent variables to the field of the rational numbers.
##  In particular, the computed regular differential chains do
##  not involve any differential polynomial which only depends
##  on the independent variables. 
##  In this package, it is, however, possible to compute decompositions of 
##  radical differential ideals generated by differential polynomials with
##  coefficients in more sophisticated differential fields.
##
##ENDSUBSECTION
##
##SUBSECTION Normal forms
##
##- Let _C_ be a regular differential chain of a differential
##  polynomial ring _R_ and _F_ be a differential polynomial of _R_.
##  Let _I_ be the differential ideal defined by _C_.
##  The *normal form* of _F_ with respect to _C_ is a rational
##  differential fraction _P_/_Q_ such that
##
##-- _Q_ is regular (i.e. a non-zero divisor) in _R_/_I_.
##
##-- _P_/_Q_ is equivalent to _F_ modulo _I_, in the sense
##   that _Q*F - P_ belongs to _I_.
##
##-- _P_ is reduced with respect to _C_, in the sense that, given
##   any leading rank _v^d_ of any element of _C_, it does not depend 
##   on any proper derivative of _v_, and, has degree less than _d_,
##   in _v_.
##
##-- _Q_ does not depend on any derivative of _v_, where _v_ denotes
##   any leading derivative of any element of _C_.
##
##- The *normal form* _P_/_Q_ of _F_ with respect to _C_ is a canonical
##  representative of the residue class of _F_ in _R_/_I_, in the 
##  sense that
##
##-- _P_/_Q_ is _0_ if and only if _F_ belongs to _I_.
##
##-- If _F_ is equivalent to _H_ modulo _I_, then, the normal forms
##   of _F_ and _H_ with respect to _C_, are equal.
##
##- The normal form of _F_ can be computed by means of the "NormalForm"
##  function [BL00].
##
##- Let us extend the above definition and consider
##  a rational differential fraction _F_/_G_, where _F_ and _G_ are
##  differential polynomials of _R_. The "NormalForm" function can
##  be used to compute a *normal form* of _F_/_G_, with respect to _C_. 
##  If it succeeds, then, it returns a rational differential fraction 
##  _P_/_Q_ such that
##
##-- _Q_ is regular (i.e. a non-zero divisor) in _R_/_I_.
##
##-- _P_/_Q_ is equivalent to _F_/_G_ modulo _I_, in the sense
##   that _Q*F - G*P_ belongs to _I_.
##
##-- _P_ is reduced with respect to _C_.
##
##-- _Q_ does not depend on any derivative of any leading derivative of _C_.
##
##- If it succeeds, then, _G_ is regular in _R_/_I_, and,
##  the *normal form* _P_/_Q_ is a canonical representative of the residue 
##  class of _F_/_G_ in the total fraction ring of _R_/_I_, in the
##  sense that
##
##-- _P_/_Q_ is _0_ if and only if _F_ belongs to _I_.
##
##-- If _F_/_G_ is equivalent to _H_/_K_ modulo _I_ (with _K_ regular
##  modulo _I_), then, the normal forms of _F_/_G_ and _H_/_K_ with respect 
##  to _C_, are equal.
##
##- The "NormalForm" function may fail to compute a *normal form* of _F_/_G_
##  if one of the three following cases occur:
##
##-- If _G_ is zero in _R_/_I_.
##
##-- If _G_ is a zero-divisor in _R_/_I_.
##
##-- If the function is led to invert another zero-divisor in _R_/_I_,
##   during the normal form computation.
##
##- However, for each rational differential fraction _F_/_G_, it is
##  possible to split _C_ into finitely many regular differential chains 
##  _C[1]_, `...`, _C[n]_, _C[n+1]_, `...`, _C[n+p]_ such that,
##  denoting _I[k]_ the radical differential ideal defined by _C[k]_
##
##-- The normal form of _F_/_G_ can be computed, with respect to _C[k]_,
##   for _1_ `<=` k `<=` _n_.
##
##-- _G_ is zero in _R_/_I[k]_, for _n+1_ `<=` _k_ `<=` _n+p_.
##
##-- The differential ideal _I_ is equal to the intersection of the 
##   differential ideals _I[k]_, for _1_ `<=` k `<=` _n+p_, 
##
##- This decomposition can be achieved by the "NormalForm" function.
##
##ENDSUBSECTION
##
##XREFMAP
##- "Examples" : Help:DifferentialAlgebra0[Examples]
##- "Glossary" : Help:DifferentialAlgebra0[Glossary]
##- "References" : Help:DifferentialAlgebra0[References]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "EssentialComponents" : Help:DifferentialAlgebra0[EssentialComponents]
##- "PowerSeriesSystem" : Help:DifferentialAlgebra0[PowerSeriesSystem]
##- "PowerSeriesSolution" : Help:DifferentialAlgebra0[PowerSeriesSolution]
##- "NormalForm" : Help:DifferentialAlgebra0[NormalForm]
##- "BelongsTo" : Help:DifferentialAlgebra0[BelongsTo]
##- "Pardi" : Help:DifferentialAlgebra0[Pardi]

