##PROCEDURE(help) DifferentialAlgebra0[Integrate]
##HALFLINE        integrates a differential rational fraction
##AUTHOR          F. Boulier
##DATE            December 2011
##
##CALLINGSEQUENCE
##- `Integrate`('p', 'x', 'R', 'opts')
##- `Integrate`('L', 'x', 'R', 'opts')
##
##PARAMETERS
##- 'p'            : a differential polynomial
##- 'x' (optional) : a variable
##- 'L'            : a list or a set of differential polynomials
##- 'R'            : a differential polynomial ring
##- 'opts' (optional) : a sequence of options
##
## iterated = true | false
## indicates if Integrate must be applied iteratively on its input

Integrate := proc (eqns,
    {iterated::boolean:=true},
    {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
    {_TIMEOUT_::nonnegative:=0},
    {_MEMOUT_::nonnegative:=0},
    {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})
    local result, leqns, innot, outnot, a_table, x;

    if _nrest <> 2 then
	error _ERRNARGS_, "Three", convert (_nrest+1, ':-string');
    elif not IsIdealDRing (_rest[_nrest]) then
	error _ERRIDRNG_, "last", convert (_rest[_nrest], ':-string');
    end if;

    x := _rest[1];

    if type (_rest[_nrest], ':-list') then
        a_table := _rest[_nrest] [1];
    else    
        a_table := _rest[_nrest];
    end if;

    try 
        leqns := ToExpandedRatfracList (eqns)
    catch:
        error _ERRPLPL_, "first", convert (eqns, ':-string') 
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    outnot := FixNotation (_OUTPUT_, _IONOT_, NotationSafe (a_table));
    try     
        result := IntegrateSafe (leqns, x, iterated, a_table, innot, outnot,
                                        _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    catch:                  
        error
    end try;    
    result := FromExpandedRatfracList (result, eqns);
    return result           
end proc:   

IntegrateSafe := proc (eqns, x, iterated, a_table, innot, outnot, 
						_TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "RatBilge"
                        (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], eqns),
                         ConvertString (a_table [':-_FREEZESEQ_'], x),
			 iterated,
                         a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, outnot, _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:

