##PROCEDURE(help) DifferentialAlgebra0[IsConstant]
##HALFLINE        tests if a rational fraction is a constant
##AUTHOR          F. Boulier
##DATE            December 2011
##
##CALLINGSEQUENCE
##- `IsConstant`('p', 'x', 'R', 'opts')
##- `IsConstant`('L', 'R', 'opts')
##
##PARAMETERS
##- 'p'            : a differential rational fraction
##- 'x' (optional) : a variable
##- 'R'            : a differential polynomial ring
##- 'opts' (optional) : a sequence of options

IsConstant := proc 
    (eqns,
     {_IONOT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_INPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_OUTPUT_::identical(:-_JET_,:-_TJET_,:-_DIFF_,:-_UDIF_,:-_U_):=':-_U_'},
     {_TIMEOUT_::nonnegative:=0},
     {_MEMOUT_::nonnegative:=0},
     {_CELLSZ_::identical(:-_SMALL_,:-_LARGE_):=':-_SMALL_'})
    local result, leqns, innot, a_table, x;

    if _nrest <> 1 and _nrest <> 2 then
	error _ERRNARGS_, "One or two", convert (_nrest+1, ':-string');
    elif not IsIdealDRing (_rest[_nrest]) then
	error _ERRIDRNG_, "last", convert (_rest[_nrest], ':-string');
    end if;
	
    if _nrest = 1 then x := 0; else x := _rest [1]; end if;
    if type (_rest[_nrest], ':-list') then
	a_table := _rest[_nrest] [1];
    else
	a_table := _rest[_nrest];
    end if;

    try
	leqns := ToExpandedRatfracList (eqns)
    catch:
	error _ERRPLPL_, "first", convert (eqns, ':-string')
    end try;
    innot := FixNotation (_INPUT_, _IONOT_, NotationSafe (a_table));
    try
	result := IsConstantSafe (leqns, x, a_table, innot,
					_TIMEOUT_, _MEMOUT_, _CELLSZ_);
    catch:
	error
    end try;
    result := FromExpandedRatfracList (result, eqns);
    return result
end proc:

IsConstantSafe := proc (eqns, x, a_table, innot, _TIMEOUT_, _MEMOUT_, _CELLSZ_)
    local arguments;

    arguments := "IsConstant"
                        (ListOrTableToBMI (a_table [':-_FREEZESEQ_'], eqns),
                         ConvertString (a_table [':-_FREEZESEQ_'], x),
                         a_table);
    arguments := arguments,
        ToBMIOptionSequence (innot, ':-_JET_', _TIMEOUT_, _MEMOUT_, _CELLSZ_);
    return BMIBladEvalSafe (a_table [':-_UNNUMBERSEQ_'], arguments);
end proc:



