##PROCEDURE(help) DifferentialAlgebra0[IsDifferentialRing]
##HALFLINE        tests if its argument is a differential ring
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- IsDifferentialRing ('R')
##
##PARAMETERS
##- 'R' : an expression
##
##DESCRIPTION
##- The function call `IsDifferentialRing`('R') returns `true` if 'R'
##  is a ::table:: encoding a polynomial differential ring, else
##  `false`.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `IsDifferentialRing(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[IsDifferentialRing](...)`.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> IsDifferentialRing(3);
##< false
##
##> R := DifferentialRing(derivations = [], blocks = []);
##
##> IsDifferentialRing(R);
##< true
##
##SEEALSO
##- "DifferentialAlgebra0", "DifferentialRing"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "DifferentialRing" : Help:DifferentialAlgebra0[DifferentialRing]


# EXPORTED
# Returns true if dring is a table representing a differential ring

IsDifferentialRing := proc (dring)
    return evalb (type (dring, ':-table') and dring [':-_TYPE_'] = ':-_DRING_')
end proc:
