##PROCEDURE(help) DifferentialAlgebra0[IsRegularDifferentialChain]
##HALFLINE        tests if its argument is a differential regular chain
##AUTHOR          F. Boulier
##DATE            August 2009
##
##CALLINGSEQUENCE
##- IsRegularDifferentialChain ('R')
##
##PARAMETERS
##- 'R' : an expression
##
##DESCRIPTION
##- The function call `IsRegularDifferentialChain`('R') returns `true` if 'R'
##  is a ::table:: encoding a regular chain of the
##  "DifferentialAlgebra0" package, else `false`.
##
##- This command is part of the `DifferentialAlgebra0` package. It can be
##  called using the form `IsRegularDifferentialChain(...)` after executing the command
##  `with`('DifferentialAlgebra0'). It can also be directly called using the
##  form `DifferentialAlgebra0[IsRegularDifferentialChain](...)`.
##
##EXAMPLES
##> with(DifferentialAlgebra0):
##> R := DifferentialRing(derivations = [t], blocks = [u]);
##
##> ideal := RosenfeldGroebner([u[t] - u], R);
##
##> IsRegularDifferentialChain(ideal[1]);
##< true
##
##> IsRegularDifferentialChain(ideal);
##< false
##
##SEEALSO
##- "DifferentialAlgebra0", "RosenfeldGroebner", "PretendRegularDifferentialChain"
##
##XREFMAP
##- "DifferentialAlgebra0" : Help:DifferentialAlgebra0
##- "Notation" : Help:DifferentialAlgebra0[Notation]
##- "RosenfeldGroebner" : Help:DifferentialAlgebra0[RosenfeldGroebner]
##- "PretendRegularDifferentialChain" : Help:DifferentialAlgebra0[PretendRegularDifferentialChain]

# EXPORTED
# Returns true if regchain is a table representing a regular chain

IsRegularDifferentialChain := proc (regchain)
    return evalb (type (regchain, ':-table') and
                    regchain [':-_TYPE_'] = ':-_REGCHAIN_');
end proc:
