# LOCAL function

# Converts a rational fraction, a list of rational fractions or 
# a set of rational fractions to a list under expanded form, 
# in order to simplify the task of the BMI parsers.
#
# See FromExpandedRatfracList which is the "converse" function
# See ToPolynomialEquationsAndInequations which performs a related task.

ToExpandedRatfracList := proc (ratfracs, {no_denom::boolean:=false})
    local result;
    if type (ratfracs, '{:-algebraic,:-equation}') then
        result := [ToExpandedRatfrac (ratfracs, no_denom)];
    elif type (ratfracs, ':-list') then
        result := map (ToExpandedRatfrac, ratfracs, no_denom);
    elif type (ratfracs, ':-set') then
        result := map (ToExpandedRatfrac, 
				convert (ratfracs, ':-list'), no_denom);
    else
        error _ERRINT_
    end if;
    return result;
end proc:

# local to ToExpandedRatfracList

ToExpandedRatfrac := proc (ratfrac0, no_denom)
    local ratfrac, num, den;
    if type (ratfrac0, ':-equation') then
	ratfrac := lhs (ratfrac0) - rhs (ratfrac0)
    else
	ratfrac := ratfrac0
    end if;
    num := expand (numer (ratfrac));
    den := expand (denom (ratfrac));
    if no_denom and not type (den, ':-numeric') then
	error
    end if;
    return num / den
end proc:
