# LOCAL function
#
# ToPolynomialEquationsAndInequations (L)
# Somehow splits L in two lists: eqns, ineqs;
#
# If p = 0 lies in L then numer (p) is moved to eqns
# If p <> 0 lies in L then numer (p) is moved to ineqs
# The denominator of p, if any, is moved to ineqs
# 
# The elements of L which do not hold any relational operator
# are considered as equations
# 
# A second list can also be provided. 
# It is processed as the first one, except that the elements which 
# do not hold any relational operator are considered as inequations.
# 
# The option 'no_ineqs' = true permits to forbid inequations.

ToPolynomialEquationsAndInequations := proc ({no_ineqs::boolean:=false})
    local eqns, ineqs, i, p, q, numq, denq;
    if _nrest < 1 or _nrest > 2 then
	error _ERRINT_
    end if;
    eqns := NULL;
    ineqs := NULL;
    for i from 1 to _nrest do
	if not type (_rest[i], {':-list',':-set'}) then
	    error _ERRINT_
	end if;
        for p in _rest[i] do
	    if type (p, `=`) or (i = 1 and type (p, ':-algebraic')) then
	        if type (p, `=`) then
		    q := lhs (p) - rhs (p);
		else
		    q := p
		end if;
	        numq := expand (numer (q));
	        denq := expand (denom (q));
	        if numq <> 0 then
		    eqns := eqns, numq
	        end if;
	        if not type (denq, ':-numeric') then
		    ineqs := ineqs, denq
	        end if
	    elif type (p, `<>`) or (i = 2 and type (p, ':-algebraic')) then
		if no_ineqs then
		    error _ERRINT_
		elif type (p, `<>`) then
		    q := lhs (p) - rhs (p);
		else
		    q := p
		end if;
	        numq := expand (numer (q));
	        denq := expand (denom (q));
	        if not (type (numq, ':-numeric') and numq <> 0) then
		    ineqs := ineqs, numq
	        end if;
	        if not type (denq, ':-numeric') then
		    ineqs := ineqs, denq
	        end if
	    else
	        error _ERRINT_
	    end if
	end do
    end do;
    return [ eqns ], [ ineqs ]
end proc:

