#test
# The Henri Michaelis Menten reduction of the basic enzymatic reaction

with (TestTools):

##read "DifferentialAlgebra0.mpl";
with (DifferentialAlgebra0):

# The basic enzymatic system
#        k1             k2                 k3
# E + S ----> C,     C ----> E + S,     C ----> E + P
#
# with the QSSA assumption: k1, k2 >> k3
#
# F. Lemaire's DAE.
# The contribution of the fast reactions is replaced by an unknown: F1(t)
# The algebraic equation states that the fast reactions are at equilibrium

Try[testnoerror] (10, [
          diff(E(t),t) = -F[1](t) + k[3]*C(t),
          diff(S(t),t) = - F[1](t),
          diff (C(t),t) = F[1](t) - k[3]*C(t),
          diff (P(t),t) = k[3]*C(t),
          k[1]*E(t)*S(t) - k[2]*C(t) = 0
        ], assign='sys');

Try[testnoerror] (20, [C,E,P,S], assign='species');

Try[testnoerror] (30, [k[1],k[2],k[3]], assign='params');

Try[testnoerror] (40, DifferentialRing (blocks = [F[1], species, params],
		notation = diff, derivations = [t], parameters = params),
		assign='R');

Try[testnoerror] (50, RosenfeldGroebner 
		(sys, [C(t)], basefield = field (generators = params), R), 
		 assign='ideal');

# The raw reduced formula

Try (60, RewriteRules (ideal[1], leader=diff(S(t),t)), 
	[diff(S(t),t) = -(E(t)*S(t)^2*k[1]^2*k[3]+E(t)*S(t)*k[1]*k[2]*k[3])/(E(t)*k[1]*k[2]+S(t)*k[1]*k[2]+k[2]^2)]);

# To get the true Henri-Michaelis-Menten formula, use extra assumptions, 
# which have nothing to do with QSSA.

Try[testnoerror] (70, [E(t) + C(t) = E[0] + C[0], S(t) + C(t) + P(t) = S[0] + C[0] + P[0]], assign='conservation_laws');

Try[testnoerror] (80, [k[1],k[2],k[3],E[0],C[0],S[0],P[0]], assign='params');

Try[testnoerror] (90, subs (P[0]=0, C[0]=0, conservation_laws),
			assign='hypotheses');

Try[testnoerror] (100, DifferentialRing (blocks = [F[1], species, params], notation = diff, derivations = [t], parameters = params), assign='R');

Try[testnoerror] (110, 
	RosenfeldGroebner 
		([ op(sys), op(hypotheses) ], R, 
		 basefield = field (generators = params)), assign='ideal');

Try[testnoerror] (120, RewriteRules (ideal[1], leader=diff(S(t),t)), assign='formula');

Try (130, formula, [diff(S(t),t) = -(S(t)^2*k[1]^2*k[3]*E[0]+S(t)*k[1]*k[2]*k[3]*E[0])/(S(t)^2*k[1]^2+2*S(t)*k[1]*k[2]+k[1]*k[2]*E[0]+k[2]^2)]);

# Some extra hypotheses and parameter renaming are necessary to recover the well-known Henri-Michaelis-Menten formula
#
# Denote K = k2/k1
# Denote Vmax = k3*E(0)
# Neglect the term K*E(0) (assuming S(0) >> E(0))

Try[testnoerror] (140, subs (k[2]=K*k[1], formula), assign='formula');
Try[testnoerror] (150, normal (formula), assign='formula');
Try[testnoerror] (160, algsubs (k[3]*E[0]=V[max], formula), assign='formula');
Try[testnoerror] (170, normal (subs (K*E[0]=0, formula)), assign='formula');

# The formula

Try (180, formula, [diff(S(t),t) = -(V[max]*S(t))/(S(t)+K)]);

#end test
