#test
# Tests the PowerSeriesSolution function

with (TestTools):

##read "DifferentialAlgebra0.mpl";
with (DifferentialAlgebra0):

# Basic use
Try[testnoerror] (10, DifferentialRing (derivations = [t], blocks = [u]), assign='R');
Try[testnoerror] (20, PretendRegularDifferentialChain ([u[t]^2-4*u], R), assign='ideal');

#################################
# ODE, one dependent variable
# The algebraic constraint is not checked. Generic series.
Try (30, PowerSeriesSolution (ideal, 4, [], check=inequations), 
				[u(t) = u(0)+eval(diff(u(t),t),{t = 0})*t+t^2]);

Try (40, PowerSeriesSolution (ideal, 2, [u=c^2,u[t]=2*c]),
				[u(t) = c^2+2*c*t+t^2]);
Try (50, PowerSeriesSolution (ideal, 4, [u=c^2,u[t]=2*c], check=all),
				[u(t) = c^2+2*c*t+t^2]);

# Change of expansion point

Try (60, PowerSeriesSolution (ideal, 2, [u=c^2,u[t]=2*c,t=t0], check=all),
				[u(t) = c^2+2*c*(t-t0)+(t-t0)^2]);

# The zero ideal

Try (70, PowerSeriesSolution (PretendRegularDifferentialChain ([], R), 2, []), [u(t) = u(0)+eval(diff(u(t),t),{t = 0})*t+1/2*eval(diff(diff(u(t),t),t),{t = 0})*t^2]);

Try[testnoerror] (80, PretendRegularDifferentialChain ([u[t]^2-t*u], R), assign='ideal');

# with an algebraic number
Try (90, PowerSeriesSolution (ideal, 3, [t=t0,u=c,u[t]=sqrt(t0*c)]), [u(t) = c+(t0*c)^(1/2)*(t-t0)+1/2*(1/2/t0*(t0*c)^(1/2)+1/2*t0)*(t-t0)^2+1/6*(3/4-1/4/t0^2*(t0*c)^(1/2))*(t-t0)^3]);

Try (100, PowerSeriesSolution (ideal, 3, [t=1,u=c^2,u[t]=c]), [u(t) = c^2+c*(t-1)+1/2*(1/2+1/2*c)*(t-1)^2+1/6*(3/4-1/4*c)*(t-1)^3]);

#################################
## ODE, two dependent variables

Try[testnoerror] (110, DifferentialRing (derivations = [t], blocks = [u,v]), assign='R');
Try[testnoerror] (120, PretendRegularDifferentialChain ([u[t]^2-4*u], R), assign='ideal');

# with an arbitrary function phi(t) for v. Just renaming v = phi
Try (130, PowerSeriesSolution (ideal, 2, [u=c^2,u[t]=2*c,v=phi(t)], check=all), [u(t) = c^2+2*c*t+t^2, v(t) = phi(0)+eval(diff(phi(t),t),{t = 0})*t+1/2*eval(diff(diff(phi(t),t),t),{t = 0})*t^2]);

# with non values for v.
Try (140, PowerSeriesSolution (ideal, 2, [u=c^2,u[t]=2*c], check=all), [u(t) = c^2+2*c*t+t^2, v(t) = v(0)+eval(diff(v(t),t),{t = 0})*t+1/2*eval(diff(diff(v(t),t),t),{t = 0})*t^2]);

# with v = 0
Try (150, PowerSeriesSolution (ideal, 2, [u=c^2,u[t]=2*c,v=0], check=all), [u(t) = c^2+2*c*t+t^2, v(t) = 0]);

# Non autonomous equation

Try[testnoerror] (160, DifferentialRing (derivations=[t], blocks = [y], notation = diff), assign='R');
Try[testnoerror] (170, diff(y(t),t)^3-4*t*y(t)*diff(y(t),t)+8*y(t)^2, assign='F');
Try[testnoerror] (180, RosenfeldGroebner([F], R), assign='ideal');

Try (190, PowerSeriesSolution (ideal[1], 3, [y(t)=c^3, diff(y(t),t)=-2*c^2]), [y(t) = c^3-2*c^2*t+c*t^2]);

#################################
# PDE, one dependent variable

Try[testnoerror] (200, DifferentialRing (derivations = [t,x], blocks = [u]), assign='R');
Try[testnoerror] (210, PretendRegularDifferentialChain ([u[t,t]-u[x]], R), assign='ideal');

# The heat equation
# The first band u(0,x) = phi(x), the second band du/dt (0,x) = psi(x)
Try (220, PowerSeriesSolution (ideal, 2, [u=phi(x),u[t]=psi(x),x=x0,t=t0]), [u(t,x) = phi(x0)+Eval(diff(phi(x),x),[x = x0])*(x-x0)+psi(x0)*(t-t0)+1/2*Eval(diff(diff(phi(x),x),x),[x = x0])*(x-x0)^2+Eval(diff(psi(x),x),[x = x0])*(t-t0)*(x-x0)+1/2*Eval(diff(phi(x),x),[x = x0])*(t-t0)^2]);

#################################
# PDE, two dependent variables

Try[testnoerror] (230, DifferentialRing (derivations = [x,y], blocks = [u,v]), assign='R');
Try[testnoerror] (240, PretendRegularDifferentialChain ([u[x,x,y], u[x,y,y]], R), assign='ideal');

# One vertical band (y axis)
# One horizontal band (x axis)
# One constant u[x,y]
#
# The vertical band du/dy(0,y) = mu(y)
# The horizontal band u(x,0)   = phi(x)
# The constant d^2u/dx*dy(0,0) = c

Try (250, PowerSeriesSolution (ideal, 3, [u=phi(x),u[y]=mu(y),u[x,y]=c,v=0]), [u(x,y) = phi(0)+mu(0)*y+eval(diff(phi(x),x),{x = 0})*x+1/2*eval(diff(mu(y),y),{y = 0})*y^2+c*x*y+1/2*eval(diff(diff(phi(x),x),x),{x = 0})*x^2+1/6*eval(diff(diff(mu(y),y),y),{y = 0})*y^3+1/6*eval(diff(diff(diff(phi(x),x),x),x),{x = 0})*x^3, v(x,y) = 0]);

Try[testnoerror] (260, v*u[x,x]^2+v*u[x,x]+u[x], assign='p1');
Try[testnoerror] (270, u[x,y], assign='p2');
Try[testnoerror] (280, u[y,y]^2-1, assign='p3');
Try[testnoerror] (290, RosenfeldGroebner ([p1,p2,p3], [u[x]], R), assign='ideal');

# First component
# One band v(x,0) = phi(x)
# Three arbitrary constants u(0,0)=c[0], ux(0,0)=0, uy(0,0)=c[1]
# One algebraic constant uyy(0,0) = -1

Try (300, PowerSeriesSolution (ideal[1], 3, [v=phi(x),u=c[0],u[x]=0,u[y]=c[1],u[y,y]=1,u[x,x]=-1], check=all), [u(x,y) = c[0]+c[1]*y+1/2*y^2-1/2*x^2-1/6/phi(0)*x^3, v(x,y) = phi(0)+eval(diff(phi(x),x),{x = 0})*x+1/2*eval(diff(diff(phi(x),x),x),{x = 0})*x^2+1/6*eval(diff(diff(diff(phi(x),x),x),x),{x = 0})*x^3]);

# Second component
# Three arbitrary constants v(0,0)=c0, u(0,0)=c1, uy(0,0)=c2
# One algebraic constant uyy(0,0) = 1
# The solutions are polynomials

Try[testnoerror] (310, PowerSeriesSolution (ideal[2], 6, [v=c0, u=c1, u[y]=c2, u[y,y]=1]), assign='sols');
Try (320, sols, [u(x,y) = c1+c2*y+1/4*c0*x+1/2*y^2-1/4*x^2, v(x,y) = c0-2*x]);

Try[testnoerror] (330, DifferentialRing (R, notation = diff), assign='S');
# Try (340, Evaluate (Equations (ideal[2], output=diff), sols, S), [0, 0, 0, 0]);
# Try (350, Evaluate (Equations ([p1,p2,p3], input=jet, S), sols, S), [0, 0, 0]);

# With the diff notation

Try[testnoerror] (340, DifferentialRing (derivations = [x,y], blocks = [u,v], notation = diff), assign='R');
Try[testnoerror] (350, [diff(diff(u(x,y),x),x)^2*v(x,y)+diff(diff(u(x,y),x),x)*v(x,y)+diff(u(x,y),x), diff(diff(u(x,y),x),y), diff(diff(u(x,y),y),y)^2-1], assign='syst');
Try[testnoerror] (360, RosenfeldGroebner (syst, [diff(u(x,y),x)], R), assign='ideal');

Try (370, PowerSeriesSolution (ideal[1], 3, [v(x,y)=f(x), u(x,y)=c[0], diff(u(x,y),x)=c[1], diff(u(x,y),y)=c[2], diff(u(x,y),y,y)=-1, diff(u(x,y),x,x)=1/2/f(x)*(-f(x)+sqrt(f(x)^2-4*f(x)*c[1]))]), [u(x,y) = c[0]+c[2]*y+c[1]*x-1/2*y^2+1/4/f(0)*(-f(0)+(f(0)^2-4*f(0)*c[1])^(1/2
))*x^2+1/6*(-1/f(0)*(-f(0)+(f(0)^2-4*f(0)*c[1])^(1/2))*c[1]*eval(diff(f(x),x),
{x = 0})/(-f(0)^2+4*f(0)*c[1])-1/2/f(0)*(-f(0)+(f(0)^2-4*f(0)*c[1])^(1/2))/(4*
c[1]-f(0))-c[1]*eval(diff(f(x),x),{x = 0})/(-f(0)^2+4*f(0)*c[1])-2*c[1]/(-f(0)
^2+4*f(0)*c[1]))*x^3, v(x,y) = f(0)+eval(diff(f(x),x),{x = 0})*x+1/2*eval(diff
(diff(f(x),x),x),{x = 0})*x^2+1/6*eval(diff(diff(diff(f(x),x),x),x),{x = 0})*x
^3]);

# With parameters

Try[testnoerror] (380, DifferentialRing (derivations = [t], blocks =[u,p], parameters = [p]), assign='R');
Try[testnoerror] (390, PretendRegularDifferentialChain ([u[t]^2-p*u], R), assign='ideal');

Try (400, PowerSeriesSolution (ideal, 10, [p=q,u=q*c^2,u[t]=q*c]), [u(t) = q*c^2 + q*c*t + 1/4*q*t^2, p = q]);

Try (410, PowerSeriesSolution (ideal, 10, [u=p*c^2,u[t]=p*c]), [u(t) = p*c^2 + p*c*t + 1/4*p*t^2, p = p]);

#end test
