#test
# Tests the PreparationEquation function

with (TestTools):

##read "DifferentialAlgebra0.mpl";
with (DifferentialAlgebra0):

# Basic use
Try[testnoerror] (10, DifferentialRing (derivations = [t], blocks = [u]), assign='R');

Try[testnoerror] (20, RosenfeldGroebner ([u[t]^2-4*u], R), assign='ideal');

Try (30, PreparationEquation (u[t]^2-4*u, ideal[2]), 
		u[t]^2-4*u = -4*z1+z1[t]^2);

Try (40, PreparationEquation (u[t]^2=4*u, ideal[2]), 
		u[t]^2-4*u = -4*z1+z1[t]^2);

# congruence
Try (50, PreparationEquation (u[t]^2-4*u, ideal[2], congruence=true),
		u[t]^2-4*u = -4*z1);

# zstring
Try (60, PreparationEquation (u[t]^2-4*u, ideal[2], zstring="toto_%d"),
		u[t]^2-4*u = -4*toto_1+toto_1[t]^2);

# rational coefficients
Try (70, PreparationEquation ((1/7)*u[t]^2-(4/13)*u^3, ideal[2]),
		1/7*u[t]^2-4/13*u^3 = 1/7*z1[t]^2-4/13*z1^3);

# With parameters
# The numbering includes the fullset of equations.

Try[testnoerror] (80, DifferentialRing (derivations = [t], blocks = [u,p], parameters = [p]), assign='R');

Try[testnoerror] (90, RosenfeldGroebner ([u-p], R), assign='ideal');

Try (100, PreparationEquation (u[t], ideal[1]), u[t] = z2[t] + z1);

Try (110, PreparationEquation (u[t], ideal[1], output=diff), 
         diff(u(t),t) = diff(z2(t),t) + z1(t));

# With base fields

Try[testnoerror] (120, DifferentialRing (derivations = [t], blocks = [u,[s,c]]), assign='R');

Try[testnoerror] (130, RosenfeldGroebner ([s[t] - c, c[t] + s], R) [1], assign='kdeal');

Try[testnoerror] (140, RosenfeldGroebner ([u[t]^2 - s*u], basefield = field (relations = kdeal), R), assign='ideal');

# First example

Try (150, PreparationEquation (s[t]*u[t], ideal[2]), s[t]*u[t] = c*z3[t]+z2*z3[t]);

Try (160, PreparationEquation (s[t]*u[t], ideal[2], congruence=true), s[t]*u[t] = c*z3[t]);

Try (170, PreparationEquation (s[t]*u[t], ideal[2], n=2), s[t]*u[t] = c*z3[t]+z2*z3[t]);

Try (180, PreparationEquation (s[t]*u[t], ideal[2], n=2, congruence=true), s[t]*u[t] = c*z3[t]);

# Second example

Try (190, PreparationEquation (s[t]*u[t] - c*u[t], ideal[2]), s[t]*u[t]-c*u[t] = z2*z3[t]);

Try (200, PreparationEquation (s[t]*u[t] - c*u[t], ideal[2], congruence=true), s[t]*u[t]-c*u[t] = z2*z3[t]);

Try (210, PreparationEquation (s[t]*u[t] - c*u[t], ideal[2], n=2), s[t]*u[t]-c*u[t] = z2*z3[t]);

Try (220, PreparationEquation (s[t]*u[t] - c*u[t], ideal[2], n=2, congruence=true), s[t]*u[t]-c*u[t] = 0);

# Third example

Try (230, PreparationEquation (s[t]*u[t]^2 + c[t] + s, ideal [2]), s[t]*u[t]^2+c[t]+s = z1+c*z3[t]^2+z2*z3[t]^2);

Try (240, PreparationEquation (s[t]*u[t]^2 + c[t] + s, ideal [2], congruence=true), s[t]*u[t]^2+c[t]+s = z1);

Try (250, PreparationEquation (s[t]*u[t]^2 + c[t] + s, ideal [2], n=2), s[t]*u[t]^2+c[t]+s = z1+c*z3[t]^2+z2*z3[t]^2);

Try (260, PreparationEquation (s[t]*u[t]^2 + c[t] + s, ideal [2], n=2, congruence=true), s[t]*u[t]^2+c[t]+s = c*z3[t]^2);

#end test
