with (LinearAlgebra):
read "/home/calforme/boulier/bmi-install/maple/DifferentialAlgebra.mpl";
with (DifferentialAlgebra);

differential := proc (p, R, d)
    local dv, v;
    dv := proc (v, R, d)
        local term, depv;
	term, depv := FactorDerivative (v, R);
	Differentiate (d(depv), term, R)
    end proc;
    add (Separant (p, v, R)*dv(v, R, d), v = Indets (p, R))
end proc:

undifferential := proc (v, R)
    local term, depv;
    term, depv := FactorDerivative (v, R);
    Differentiate (op (depv), term, R)
end:

enlargeMatrix := proc (M, indices::list, expr, ideal::table, derv)
    local newM, new_indices, l, c, NF, tc, i, k, koeff, terme, koeffs, termes;

    c := ColumnDimension (M);
    newM := <Copy (M), Vector[row] (c, 0)>;
    l := RowDimension (newM);
    newM [l, c] := expr;

    NF := NormalForm (expr, ideal);
    koeffs, termes := Coeffs (NF, derv, ideal);

    new_indices := indices;
    for i from 1 to nops (termes) do
	koeff := koeffs [i];
	terme := termes [i];
	if member (terme, new_indices, 'k') then
	    newM [l, k] := koeff
	else
	    newM := < newM [1 .. l, 1 .. c-1] | 
		      Vector[column] (l, 0) |
		      newM [1 .. l, c] >;
	    newM [l, c] := koeff;
	    c := c + 1;
	    new_indices := [ op (new_indices), terme ]
	end if
    end do;
    newM, new_indices
end proc:

alea := rand (-100..100);

enlargeMatrixProb := proc (M, indices::list, expr, 
				variables::list,
				values::list, 
				ideal::table, derv)
    local newM, new_indices, new_variables, new_values, 
		new_variables2, new_values2, koeffs, termes,
		l, c, NF, tc, i, j, k, z, X, koeff, terme;

    c := ColumnDimension (M);
    newM := <Copy (M), Vector[row] (c, 0)>;
    l := RowDimension (newM);
    newM [l, c] := expr;

    NF := NormalForm (expr, ideal);
    koeffs, termes := Coeffs (NF, derv, ideal);

    new_indices := indices;
    new_variables := variables;
    new_values := values;
    for i from 1 to nops (termes) do
        koeff := koeffs[i];
	X := Indets (koeff, ideal);
	X := SortByRank (X, descending, ideal);
	for j from 1 to nops (X) do
	    if not member (X[j], new_variables, 'k') then
		new_variables := [ op(new_variables), X[j] ];
		do
		    z := alea ();
		    if z <> 0 then break end if
		od;
		new_values := [ op(new_values), z ]
	    end if
	end do;
	new_variables2 := SortByRank (new_variables, descending, ideal);
	new_values2 := [];
	for j from 1 to nops (new_variables2) do
	    member (new_variables2 [j], new_variables, 'k');
	    new_values2 := [ op (new_values2), new_values [k] ]
	end do;
	new_variables := new_variables2;
	new_values := new_values2;
	z := subs (seq (new_variables[j] = new_values[j], 
				j = 1 .. nops (new_variables)), koeff);
	koeff := z;
        terme := termes [i];
        if member (terme, new_indices, 'k') then
            newM [l, k] := koeff
        else
            newM := < newM [1 .. l, 1 .. c-1] |
                      Vector[column] (l, 0) |
                      newM [1 .. l, c] >;
            newM [l, c] := koeff;
            c := c + 1;
            new_indices := [ op (new_indices), terme ]
        end if
    end do;
    newM, new_indices, new_variables, new_values
end proc:

pivote := proc (M)
    local P, L, U, l, c;

    l := RowDimension (M);
    c := ColumnDimension (M);
    P, L, U := LUDecomposition (M [1 .. l, 1 .. c-1]);
    L^(-1) . P^(-1) . M;
end proc:

deplin := proc (M)
    local P, l, c, res, i, j, found;

    P := pivote (M);
    l := RowDimension (M);
    c := ColumnDimension (M);
    res := NULL;
    for i from 1 to l do 
	found := false;
        for j from 1 to c-1 while not found do
	    if P [i,j] <> 0 then
		found := true
	    end if
        end do;
	if not found then
	    res := res, P [i,c]
	end if
    end do;
    [res]
end proc:

allmons2 := proc (m, X, deg, R)
    local sol, x, i, Y;
    if degree (m) = deg then
	sol := m
    else
	if degree (m) = 0 then
	    Y := X
	else
	    x := MinRankElement (Indets (m, selection = allvars, R), R);
	    Y := [Equations (X, rank <= x, R)];
	end if;
	sol := NULL;
	for i from 1 to nops (Y) do
	    sol := sol, allmons2 (m*Y[i], X, deg, R)
	end do
    end if;
    sol
end proc:

allmons := proc (X, deg, R)
    local sol;
    sol := [allmons2 (1, X, deg, R)];
    sol := SortByRank (sol, ascending, R);
    sol
end proc:

allders := proc (X, ordre, R, {stairs:={}})
    local ders, M, sol, i, j, v, d;
    ders := subs (op (Ordering (R)), derivations);
    M := allmons (ders, ordre, R);
    sol := NULL;
    for i from 1 to nops (M) do
	for j from 1 to nops (X) do
	    v := X[j];
	    for d in Indets (M [i], selection = allvars, R) do
		v := Differentiate (v, d $ degree (M[i], d), R)
	    od;
	    sol := sol, v
	od
    od;
    for i from 1 to nops (stairs) do
	sol := Equations ([sol], leader <> derivative(stairs[i]), R)
    od;
    sol := SortByRank ([sol], ascending, R);
    sol
end proc:

