# read "/export/boulier/bmi-install/maple/DifferentialAlgebra.mpl":
# with (DifferentialAlgebra):

# R := DifferentialRing (derivations = [x], blocks = [u], notation = tjet);

# Returns A and whatsleft such that
# F = A' + whatsleft

RatBilge := proc (F, x, R, whatsleft)
    return RatBilge2 (F, x, R, whatsleft, true)
end proc:

RatBilge2 := proc (F0, x, R, whatsleft, doquo)
    local F, P, Q, a, S, aBn_1, aBn, i;
    F := normal (F0);
    P := expand (numer (F));
    Q := expand (denom (F));
    if type (Q, numeric) then
	return PolBilge (F, x, R, whatsleft)
    end if;
    S := sqrfree (Q);
    a := S [1];
    S := S [2];
    for i from 1 to nops (S) do
	if S [i,2] = 1 then
	    P := S [i,1]*P;
	    Q := S [i,1]*Q;
	    S [i,2] := 2
	end if
    end do;
# This test should always fail
    if select (pair -> pair [2] = 1, S) <> [] then
	error "this test should always fail";
	whatsleft := F;
	return 0;
    end if;
    aBn_1 := a*mul (S [i,1]^(S [i,2]-2), i = 1 .. nops (S));
    aBn := a*mul (S [i,1]^(S [i,2]-1), i = 1 .. nops (S));
    return BilgeRat (expand (aBn_1*P), aBn, x, R, whatsleft, doquo);
end proc:

# Returns A, whatsleft such that
#                          N
#                         ---- = A' + whatsleft
#                           2
#                          B
# B is assumed not numeric

BilgeRat := proc (N, B, x, R, whatsleft, doquo_0)
    local doquo, toprocess, processed, notprocessed, new_toprocess,
	mbar, new_mbar, vbar, dbar, cbar, new_cbar, d, theta, u,
	Bx, mBx, vBx, cBx, 
	mB, vB, dB, cB, 
	mA, vA, cA, tA, dtA,
	q, r, Aq, Ar, wq;

userinfo (3, RatBilge, "Entering RatBilge");
userinfo (4, RatBilge, N);

    toprocess := expand (N);
    processed := 0;
    notprocessed := 0;

    mB := LeadingMonomial (B, R, 'cB');
    vB := LeadingDerivative (mB, R);
    dB := degree (mB, vB);
    cB := cB;

    Bx := Differentiate (B, x, R);
    mBx := LeadingMonomial (Bx, R, 'cBx');
    vBx := LeadingDerivative (Bx, R);
    cBx := cBx;

    doquo := doquo_0;
    while toprocess <> 0 do
userinfo (3, RatBilge, "next loop");
        mbar := LeadingMonomial (toprocess, R, 'cbar');
        cbar := cbar;

userinfo (3, RatBilge, "mbar" = mbar);

	if type (mbar, numeric) then
            notprocessed := notprocessed + cbar*mbar;
            toprocess := toprocess - cbar*mbar;
	    next;
	end if;

	vbar := LeadingDerivative (mbar, R);
	dbar := degree (mbar, vbar);
	theta, u := FactorDerivative (vbar, R);

	if MaxRankElement ([vbar, vBx], R) <> vBx then
# Read: if vbar > vBx then mbar must come from A'B
	    if dbar > 1 or degree (theta, x) = 0 then
		notprocessed := notprocessed + cbar*mbar;
		toprocess := toprocess - cbar*mbar;
		next
	    end if;
	    vA := Differentiate (u, theta/x, R);
	    if not divide (mbar * vA, vbar * mB, 'mA') then
		notprocessed := notprocessed + cbar*mbar;
		toprocess := toprocess - cbar*mbar;
		next
	    end if;
	    d := degree (mbar, vA) + 1;
	    cA := cbar / (d * cB);
	else
# If this test applies then the loop is broken (return statement)
 	    d := degree (toprocess, vB);
	    if d >= 2*dB and doquo then
userinfo (3, BilgeRat, "perform the quotient by B^2");
	        q := quo (toprocess, B^2, vB, 'r');
# lprint ("q= ", q);
		r := r;
	        Aq := RatBilge2 (q, x, R, 'wq', true);
	        Ar := RatBilge2 (wq + r/B^2, x, R, 'wr', false);
		whatsleft := wr + notprocessed/B^2;
	        return processed/B + Aq + Ar
	    end if;

	    if not divide (mbar, mBx, 'mA') then
# Read: if lm(B') does not divide mbar then
# One assumes that mbar = lm (A'B) 
userinfo (3, BilgeRat, "mBx does not divide mbar", mBx, mbar);
	        if degree (theta, x) = 0 then
		    notprocessed := notprocessed + cbar*mbar;
		    toprocess := toprocess - cbar*mbar;
		    next
		end if;
		vA := Differentiate (u, theta/x, R);
	        d := degree (mbar, vA) + 1;
	        if not divide (mbar * vA, vbar * mB, 'mA') then
		    notprocessed := notprocessed + cbar*mbar;
		    toprocess := toprocess - cbar*mbar;
		    next
	        end if;
	        cA := cbar / (d * cB);
	    else
userinfo (3, BilgeRat, "mBx = mA * mbar", mBx, mA, mbar);
# Read: if lm(B') = mA * mBx then
		if mA <> 1 then 
		    vA := LeadingDerivative (mA, R);
		    d := degree (mA, vA)
		end if;
		if mA = 1 or vA <> Differentiate (u, theta/x, R) then
userinfo (3, BilgeRat, "mA has lower leader than (theta/x)u", 
				mA, Differentiate (u, theta/x, R));
		    cA := - cbar / cBx;
		else
userinfo (3, BilgeRat, "mA has the same leader as (theta/x)u",
				mA, Differentiate (u, theta/x, R)); 
		    if d*cB - cBx = 0 then
userinfo (3, BilgeRat, "avoid division by zero", d, cB, cBx);
			notprocessed := notprocessed + cbar*mbar;
			toprocess := toprocess - cbar*mbar;
			next
		    end if;
		    cA := cbar / (d*cB - cBx);
		end if
	    end if
	end if;
# cA and mA are computed.
# Let us proceed
	tA := cA*mA;
	dtA := Differentiate (tA, x, R);
	new_toprocess := expand (toprocess - dtA*B + tA*Bx);
	if new_toprocess <> 0 then
	    new_mbar := LeadingMonomial (new_toprocess, R, 'new_cbar');
	end if;
	if new_toprocess <> 0 and 
		(new_mbar = mbar or
		MaxRankElement ([mbar, new_mbar], R) <> mbar) then
	    notprocessed := notprocessed + cbar*mbar;
	    toprocess := toprocess - cbar*mbar
	else
	    processed := processed + tA;
	    toprocess := new_toprocess;
	    doquo := true
	end if
    end do;
userinfo (3, BilgeRat, "Final test");
    if normal (N/B^2 - Differentiate (processed/B, x, R) - 
					notprocessed/B^2) <> 0 then
	error "wrong output"
    end if;
# lprint ("");
# lprint (notprocessed/B^2);
# lprint ("");
    whatsleft := notprocessed/B^2;
    return processed/B;
end proc:

# Returns A and whatsleft such that
# pbar is a polynomial
# pbar = A' + whatsleft

PolBilge := proc (pbar, t, R, whatsleft)
    local toprocess, processed, notprocessed,
          abar, vbar, dbar, qbar, rbar,
          a, v, d, q, da,
          rk, theta, u;
    toprocess := expand (pbar);
    processed:= 0;
    notprocessed := 0;
# Loop invariant: d/dt (processed) + notprocessed + toprocess = pbar
    while toprocess <> 0 do
# The case of constant polynomials
        rk := LeadingRank (toprocess, R);
        if rk = 0 or rk = 1 then
            processed := processed + toprocess * t;
            toprocess := 0;
        else
# Denote pbar = abar * vbar^dbar + qbar
            vbar := LeadingDerivative (rk, R);
            theta, u := FactorDerivative (vbar, R);
            dbar := degree (rk, vbar);
            abar := Initial (toprocess, R);
            qbar := Tail (toprocess, R);
            if dbar > 1 or degree (theta, t) = 0 then
                notprocessed := notprocessed + abar * vbar^dbar;
                toprocess := qbar;
            else
# We look for p such that dp/dt = pbar
#                         p     = a*v^d + q
                v := Differentiate (u, theta/t, R);
                d := degree (abar, v) + 1;
                a := quo (abar, v^(d-1), v, 'rbar')/d;
# Denote da = da/dt. The rank of da is different from the one of vbar.
# It must be lower than that of vbar.
                da := Differentiate (a, t, R);
                if MaxRankElement ([da, vbar], R) = da then
                    notprocessed := notprocessed + abar * vbar^dbar;
                    toprocess := qbar;
                else
                    processed := processed + a*v^d;
                    toprocess := qbar + rbar*vbar - da*v^d
                end if
            end if
        end if
    end do;
# lprint ("");
# lprint (notprocessed);
# lprint ("");
    whatsleft := notprocessed;
    return processed;
end proc:

######################################################################
# SUBFUNCTIONS
######################################################################

# Returns the leading monomial of R.
# The numerical leading coefficient is returned in koeff.

LeadingMonomial := proc (pp, R, koeff)
    local p, m, rk;
    p := Differentiate (pp, 1, R);
    m := 1;
    do
        rk := LeadingRank (p, R);
        if rk = 0 then
            error
        elif rk = 1 then
            koeff := p;
            return m;
        else
            m := m * rk;
            p := Initial (p, R)
        end if
    end do
end proc:

######################################################################
# Tests
######################################################################

RandomPoly := proc (X, maxnorm, maxdeg)
    local x, d, i;
    if X = {} then
        return rand (-maxnorm .. maxnorm)()
    else
        x := X [rand (1 .. nops (X))()];
        d := rand (0 .. maxdeg)();
        return add (RandomPoly (X minus { x }, maxnorm, maxdeg)*x^i, i = 0 .. d)    end if
end proc:

RandomExample := proc (v, d, X, maxnorm, maxdeg)
    local A, B, i;
    A := add (RandomPoly (X, maxnorm, maxdeg)*v^i, i = 0 .. d);
    B := add (RandomPoly (X, maxnorm, maxdeg)*v^i, i = 0 .. d);
    return A, B
end proc:

RunTests := proc (R)
    global u, x, A, B, drat, rat;
    local i, aa, bb;
    for i to 10 do
#        A, B := RandomExample (u[x,x], 3, {u[x]}, 3, 2);
	A, B := RandomExample (u[x,x], 2, {u[x],u[]}, 3, 2);
        lprint ("A" = A, "B" = B);
        rat := A/B;
        drat := Differentiate (A/B, x, R);
        aa := RatBilge (drat, x, R, 'bb');
        if bb <> 0 then
            error "failure in RunTests"
        end if;
        if normal (drat - Differentiate (aa, x, R)) <> 0 then
            error
	else
	    lprint ("success")
        end if;
    end do;
end proc:

######################################################################
# Examples

# A := 1+(-2+(-3+u[x])*u[])*u[x,x]+(3-2*u[x]+2*u[])*u[x,x]^2:
# B := -1-3*u[]+2*u[x]+(2*u[]+3*u[]^2)*u[x]^2+(1+(-2+2*u[x])*u[])*u[x,x]:

# A := v*u[x]^2-v[x]*u*u[x]+1;
# B := w*u[x]^2-w[x]*u*u[x]+1;

Reformulate := proc (relES, R)
    local S, p, w, res, n;
    S := DifferentialRing (R, notation = tjet);
    p := Differentiate (relES, 1, R, output = tjet);
    res := 0;
    n := 0;
    while p <> 0 do
	p := RatBilge (p, t, S, 'w');
	w := Differentiate (w, 1, S, output = Notation (R));
	if n = 0 then
	    res := res + w;
	else
	    res := res + Diff (w, t$n);
	end if;
	n := n + 1
    end do;
    res
end proc:

