#!/bin/bash
set -e -u -x

function repair_wheel {
    wheel="$1"
    if ! auditwheel show "$wheel"; then
        echo "Skipping non-platform wheel $wheel"
    else
        auditwheel repair "$wheel" --plat "$PLAT" -w /io/dist/
    fi
}

# the versions of python
pyvers="38 39 310 311 312"

echo BUILD WHEELS
for pyver in ${pyvers} ; do
    pybin=/opt/python/cp${pyver}-cp${pyver}/bin
    echo "--------------------------------------------------"
    echo "pybin = ${pybin}"
    "${pybin}/pip" install -r /io/dev-requirements.txt
    "${pybin}/pip" wheel /io/ --no-deps -w dist/
done

# Repair wheels
echo REPAIR WHEELS
for whl in dist/*.whl; do
    echo "--------------------------------------------------"
    echo "whl = ${whl}"
    repair_wheel "$whl"
done

# Test wheels
echo TEST WHEELS
for pyver in ${pyvers} ; do
    pybin=/opt/python/cp${pyver}-cp${pyver}/bin
    echo "--------------------------------------------------"
    echo "pybin = ${pybin}"
    "${pybin}/pip" install DifferentialAlgebra --no-index -f /io/dist
    (cd /io/tests ; "${pybin}/python" -m unittest *.py)
done

